"""Functions for polygon geometries."""

from collections.abc import Callable

import networkx as nx
import numpy as np
import pandas as pd
from geopandas import GeoDataFrame
from geopandas import GeoSeries
from shapely import STRtree
from shapely import area
from shapely import box
from shapely import buffer
from shapely import difference
from shapely import extract_unique_points
from shapely import get_exterior_ring
from shapely import get_interior_ring
from shapely import get_num_geometries
from shapely import get_num_interior_rings
from shapely import get_parts
from shapely import is_empty
from shapely import make_valid
from shapely import polygons
from shapely import union_all
from shapely.errors import GEOSException
from shapely.geometry import LinearRing
from shapely.ops import SplitOp

from ..debug_config import _DEBUG_CONFIG
from ..debug_config import _try_debug_print
from ..maps.maps import explore_locals
from .conversion import to_gdf
from .conversion import to_geoseries
from .duplicates import _get_intersecting_geometries
from .general import _grouped_unary_union
from .general import _parallel_unary_union
from .general import _parallel_unary_union_geoseries
from .general import _push_geom_col
from .general import _unary_union_for_notna
from .general import clean_geoms
from .general import extend_lines
from .general import get_grouped_centroids
from .general import get_line_segments
from .general import to_lines
from .geometry_types import get_geom_type
from .geometry_types import make_all_singlepart
from .geometry_types import to_single_geom_type
from .neighbors import get_neighbor_indices
from .overlay import _try_difference
from .overlay import clean_overlay
from .polygons_as_rings import PolygonsAsRings
from .sfilter import sfilter
from .sfilter import sfilter_inverse

PRECISION = 1e-3
_BUFFER = False


def get_polygon_clusters(
    *gdfs: GeoDataFrame | GeoSeries,
    cluster_col: str = "cluster",
    allow_multipart: bool = False,
    predicate: str | None = "intersects",
    as_string: bool = False,
) -> GeoDataFrame | tuple[GeoDataFrame]:
    """Find which polygons overlap without dissolving.

    Devides polygons into clusters in a fast and precice manner by using spatial join
    and networkx to find the connected components, i.e. overlapping geometries.
    If multiple GeoDataFrames are given, the clusters will be based on all
    combined.

    This can be used instead of dissolve+explode, or before dissolving by the cluster
    column. This has been tested to be a lot faster if there are many
    non-overlapping polygons, but somewhat slower than dissolve+explode if most
    polygons overlap.

    Args:
        gdfs: One or more GeoDataFrames of polygons.
        cluster_col: Name of the resulting cluster column.
        allow_multipart: Whether to allow mutipart geometries in the gdfs.
            Defaults to False to avoid confusing results.
        predicate: Spatial predicate. Defaults to "intersects".
        as_string: Whether to return the cluster column values as a string with x and y
            coordinates. Convinient to always get unique ids.
            Defaults to False because of speed.

    Returns:
        One or more GeoDataFrames (same amount as was given) with a new cluster column.

    Examples:
    ---------
    Create geometries with three clusters of overlapping polygons.

    >>> import sgis as sg
    >>> gdf = sg.to_gdf([(0, 0), (1, 1), (0, 1), (4, 4), (4, 3), (7, 7)])
    >>> buffered = sg.buff(gdf, 1)
    >>> gdf
                                                geometry
    0  POLYGON ((1.00000 0.00000, 0.99951 -0.03141, 0...
    1  POLYGON ((2.00000 1.00000, 1.99951 0.96859, 1....
    2  POLYGON ((1.00000 1.00000, 0.99951 0.96859, 0....
    3  POLYGON ((5.00000 4.00000, 4.99951 3.96859, 4....
    4  POLYGON ((5.00000 3.00000, 4.99951 2.96859, 4....
    5  POLYGON ((8.00000 7.00000, 7.99951 6.96859, 7....

    Add a cluster column to the GeoDataFrame:

    >>> gdf = sg.get_polygon_clusters(gdf, cluster_col="cluster")
    >>> gdf
       cluster                                           geometry
    0        0  POLYGON ((1.00000 0.00000, 0.99951 -0.03141, 0...
    1        0  POLYGON ((2.00000 1.00000, 1.99951 0.96859, 1....
    2        0  POLYGON ((1.00000 1.00000, 0.99951 0.96859, 0....
    3        1  POLYGON ((5.00000 4.00000, 4.99951 3.96859, 4....
    4        1  POLYGON ((5.00000 3.00000, 4.99951 2.96859, 4....
    5        2  POLYGON ((8.00000 7.00000, 7.99951 6.96859, 7....

    If multiple GeoDataFrames are given, all are returned with common
    cluster values.

    >>> gdf2 = sg.to_gdf([(0, 0), (7, 7)])
    >>> gdf, gdf2 = sg.get_polygon_clusters(gdf, gdf2, cluster_col="cluster")
    >>> gdf2
       cluster                 geometry
    0        0  POINT (0.00000 0.00000)
    1        2  POINT (7.00000 7.00000)
    >>> gdf
       cluster                                           geometry
    0        0  POLYGON ((1.00000 0.00000, 0.99951 -0.03141, 0...
    1        0  POLYGON ((2.00000 1.00000, 1.99951 0.96859, 1....
    2        0  POLYGON ((1.00000 1.00000, 0.99951 0.96859, 0....
    3        1  POLYGON ((5.00000 4.00000, 4.99951 3.96859, 4....
    4        1  POLYGON ((5.00000 3.00000, 4.99951 2.96859, 4....
    5        2  POLYGON ((8.00000 7.00000, 7.99951 6.96859, 7....

    Dissolving 'by' the cluster column will make the dissolve much
    faster if there are a lot of non-overlapping polygons.

    >>> dissolved = gdf.dissolve(by="cluster", as_index=False)
    >>> dissolved
       cluster                                           geometry
    0        0  POLYGON ((0.99951 -0.03141, 0.99803 -0.06279, ...
    1        1  POLYGON ((4.99951 2.96859, 4.99803 2.93721, 4....
    2        2  POLYGON ((8.00000 7.00000, 7.99951 6.96859, 7....
    """
    if isinstance(gdfs[-1], str):
        *gdfs, cluster_col = gdfs

    concated = []
    orig_indices = ()

    # take a copy only if there are gdfs with the same id
    # To not get any overwriting in the for loop
    if sum(df1 is df2 for df1 in gdfs for df2 in gdfs) > len(gdfs):
        gdfs = [gdf.copy() for gdf in gdfs]

    for i, gdf in enumerate(gdfs):
        if isinstance(gdf, GeoSeries):
            gdf = gdf.to_frame()

        if not isinstance(gdf, GeoDataFrame):
            raise TypeError("'gdfs' should be GeoDataFrames or GeoSeries.")

        if not allow_multipart and len(gdf) != len(gdf.explode(index_parts=False)):
            raise ValueError(
                "All geometries should be exploded to singlepart "
                "in order to get correct polygon clusters. "
                "To allow multipart geometries, set allow_multipart=True"
            )

        orig_indices = orig_indices + (gdf.index,)

        gdf = gdf.assign(i__=i)

        concated.append(gdf)

    concated = pd.concat(concated, ignore_index=True)

    if not len(concated):
        return concated.drop("i__", axis=1).assign(**{cluster_col: []})

    concated[cluster_col] = get_cluster_mapper(concated, predicate)

    if as_string:
        concated[cluster_col] = get_grouped_centroids(concated, groupby=cluster_col)

    concated = _push_geom_col(concated)

    n_gdfs = concated["i__"].unique()

    if len(n_gdfs) == 1:
        concated.index = orig_indices[0]
        return concated.drop(["i__"], axis=1)

    unconcated = ()
    for i in n_gdfs:
        gdf = concated[concated["i__"] == i]
        gdf.index = orig_indices[i]
        gdf = gdf.drop(["i__"], axis=1)
        unconcated = unconcated + (gdf,)

    return unconcated


def get_cluster_mapper(
    gdf: GeoDataFrame | GeoSeries, predicate: str = "intersects"
) -> dict[int, int]:
    if not gdf.index.is_unique:
        raise ValueError("Index must be unique")
    neighbors = get_neighbor_indices(gdf, gdf, predicate=predicate)

    edges = [(source, target) for source, target in neighbors.items()]

    graph = nx.Graph()
    graph.add_edges_from(edges)

    return {
        j: i
        for i, component in enumerate(nx.connected_components(graph))
        for j in component
    }


def eliminate_by_longest(
    gdf: GeoDataFrame | tuple[GeoDataFrame],
    to_eliminate: GeoDataFrame,
    *,
    fix_double: bool = True,
    ignore_index: bool = False,
    aggfunc: str | dict | list | None = None,
    grid_size=None,
    n_jobs: int = 1,
    **kwargs,
) -> tuple[GeoDataFrame]:
    """Dissolves selected polygons with the longest bordering neighbor polygon.

    Eliminates selected geometries by dissolving them with the neighboring
    polygon with the longest shared border. The index and column values of the
    large polygons will be kept, unless else is specified.

    Note that this might be a lot slower than eliminate_by_largest.

    Args:
        gdf: GeoDataFrame with polygon geometries, or a list of GeoDataFrames.
        to_eliminate: The geometries to be eliminated by 'gdf'.
        fix_double: If True, geometries to be eliminated will be erased by overlapping
            geometries to not get double surfaces if the geometries in 'to_eliminate'
            overlaps with multiple geometries in 'gdf'.
        ignore_index: If False (default), the resulting GeoDataFrame will keep the
            index of the large polygons. If True, the resulting axis will be labeled
            0, 1, …, n - 1.
        aggfunc: Aggregation function(s) to use when dissolving/eliminating.
            Defaults to None, meaning the values of 'gdf' is used. Otherwise,
            aggfunc will be passed to pandas groupby.agg. note: The geometries of
            'gdf' are sorted first, but if 'gdf' has missing values, the resulting
            polygons might get values from the polygons to be eliminated
            (if aggfunc="first").
        grid_size: Rounding of the coordinates. Defaults to None.
        n_jobs: Number of threads to use. Defaults to 1.
        **kwargs: Keyword arguments passed to the dissolve method.

    Returns:
        A tuple of the GeoDataFrame with the geometries of 'to_eliminate'
        dissolved in and a GeoDataFrame with the potentionally isolated
        polygons that could not be eliminated. If multiple GeoDataFrame
        are passed as 'gdf', the returned tuple will contain each frame
        plus the isolated polygons as the last item.

    Examples:
    ---------
    Create two polygons with a sliver in between:

    >>> sliver = sg.to_gdf(Polygon([(0, 0), (0.1, 1), (0, 2), (-0.1, 1)]))
    >>> small_poly = sg.to_gdf(
    ...     Polygon([(0, 0), (-0.1, 1), (0, 2), (-1, 2), (-2, 2), (-1, 1)])
    ... )
    >>> large_poly = sg.to_gdf(
    ...     Polygon([(0, 0), (0.1, 1), (1, 2), (2, 2), (3, 2), (3, 0)])
    ... )

    Using multiple GeoDataFrame as input, the sliver is eliminated into the small
    polygon (because it has the longest border with sliver).

    >>> small_poly_eliminated, large_poly_eliminated, isolated = sg.eliminate_by_longest(
    ...     [small_poly, large_poly], sliver
    ... )

    With only one input GeoDataFrame:

    >>> polys = pd.concat([small_poly, large_poly])
    >>> eliminated, isolated = sg.eliminate_by_longest(polys, sliver)
    """
    _recurse = kwargs.pop("_recurse", False)

    if not len(to_eliminate) or not len(gdf):
        if isinstance(gdf, (list, tuple)):
            return (*gdf, to_eliminate)
        return gdf, to_eliminate

    if isinstance(gdf, (list, tuple)):
        # concat, then break up the dataframes in the end
        was_multiple_gdfs = True
        original_cols = [df.columns for df in gdf]
        gdf = pd.concat(df.assign(**{"_df_idx": i}) for i, df in enumerate(gdf))
    else:
        was_multiple_gdfs = False

    crs = gdf.crs
    geom_type = get_geom_type(gdf)

    if not ignore_index:
        idx_mapper = dict(enumerate(gdf.index))
        idx_name = gdf.index.name

    gdf = gdf.reset_index(drop=True)

    # TODO: is it ok to singlepart here?
    gdf = make_all_singlepart(gdf, ignore_index=True).pipe(
        to_single_geom_type, "polygon"
    )

    if _BUFFER:
        gdf.geometry = gdf.buffer(
            PRECISION,
            resolution=1,
            join_style=2,
        )

        to_eliminate.geometry = to_eliminate.buffer(
            PRECISION,
            resolution=1,
            join_style=2,
        )

    # more_than_one = get_num_geometries(to_eliminate.geometry.values) > 1
    # filt = more_than_one, to_eliminate._geometry_column_name
    # to_eliminate.loc[*filt] = to_eliminate.loc[*filt].apply(_unary_union_for_notna)

    gdf["_dissolve_idx"] = gdf.index
    to_eliminate = to_eliminate.assign(_eliminate_idx=lambda x: range(len(x)))

    # convert to lines to get the borders
    lines_eliminate = to_lines(to_eliminate[["_eliminate_idx", "geometry"]])

    borders = clean_overlay(
        gdf[["_dissolve_idx", "geometry"]],
        lines_eliminate,
        keep_geom_type=False,
        grid_size=grid_size,
        n_jobs=n_jobs,
    ).loc[lambda x: x["_eliminate_idx"].notna()]

    borders["_length"] = borders.length

    # as DataFrame because GeoDataFrame constructor is expensive
    borders = pd.DataFrame(borders)

    longest_border = borders.sort_values("_length", ascending=False).drop_duplicates(
        "_eliminate_idx"
    )

    to_dissolve_idx = longest_border.set_index("_eliminate_idx")["_dissolve_idx"]
    to_eliminate["_dissolve_idx"] = to_eliminate["_eliminate_idx"].map(to_dissolve_idx)

    actually_eliminate = to_eliminate.loc[to_eliminate["_dissolve_idx"].notna()]

    isolated = to_eliminate.loc[to_eliminate["_dissolve_idx"].isna()]
    containing_eliminators = (
        pd.DataFrame(
            isolated.drop(columns="_dissolve_idx").sjoin(
                gdf[["_dissolve_idx", "geometry"]], predicate="contains"
            )
        )
        .drop(columns="index_right")
        .drop_duplicates("_eliminate_idx")
    )
    isolated = isolated.drop(
        ["_dissolve_idx", "_length", "_eliminate_idx"],
        axis=1,
        errors="ignore",
    )

    eliminated = _eliminate(
        pd.DataFrame(gdf),
        pd.concat([actually_eliminate, containing_eliminators]),
        aggfunc,
        crs,
        fix_double,
        grid_size=grid_size,
        n_jobs=n_jobs,
        **kwargs,
    )

    if not ignore_index:
        eliminated.index = eliminated.index.map(idx_mapper)
        eliminated.index.name = idx_name

    eliminated = eliminated.drop(
        ["_dissolve_idx", "_length", "_eliminate_idx"],
        axis=1,
        errors="ignore",
    )

    out = GeoDataFrame(eliminated, geometry="geometry", crs=crs).pipe(clean_geoms)

    if _BUFFER:
        out.geometry = out.buffer(
            -PRECISION,
            resolution=1,
            join_style=2,
        )
        isolated.geometry = isolated.buffer(
            -PRECISION,
            resolution=1,
            join_style=2,
        )

    if geom_type != "mixed":
        out = to_single_geom_type(out, geom_type)

    out = out.reset_index(drop=True) if ignore_index else out

    _try_debug_print("inni eliminate_by_longest")
    explore_locals(center=_DEBUG_CONFIG["center"])

    if not _recurse and len(isolated):
        if 0:
            isolated.geometry = isolated.buffer(
                -PRECISION,
                resolution=1,
                join_style=2,
            )
        out, isolated = _recursively_eliminate_new_neighbors(
            out,
            isolated,
            func=eliminate_by_longest,
            fix_double=fix_double,
            ignore_index=ignore_index,
            aggfunc=aggfunc,
            grid_size=grid_size,
            n_jobs=n_jobs,
        )

    _try_debug_print("inni eliminate_by_longest 2")
    explore_locals(center=_DEBUG_CONFIG["center"])

    # assert (
    #     out[["ARTYPE", "ARTRESLAG", "ARSKOGBON", "ARGRUNNF", "kilde"]]
    #     .notna()
    #     .all()
    #     .all()
    # ), out[["ARTYPE", "ARTRESLAG", "ARSKOGBON", "ARGRUNNF", "kilde"]].sort_values(
    #     ["ARTYPE", "ARTRESLAG", "ARSKOGBON", "ARGRUNNF", "kilde"]
    # )

    if not was_multiple_gdfs:
        return out, isolated

    gdfs = ()
    for i, cols in enumerate(original_cols):
        df = out.loc[out["_df_idx"] == i, cols]
        gdfs += (df,)

    return (*gdfs, isolated)


def _recursively_eliminate_new_neighbors(
    df: GeoDataFrame,
    isolated: GeoDataFrame,
    func: Callable,
    **kwargs,
):
    len_now = len(isolated)
    while len(isolated):
        _try_debug_print(f"recurse len({len(isolated)})")
        df, isolated = func(
            df,
            isolated,
            _recurse=True,
            **kwargs,
        )

        if len_now == len(isolated):
            break
        len_now = len(isolated)

    return df, isolated


def eliminate_by_largest(
    gdf: GeoDataFrame | list[GeoDataFrame],
    to_eliminate: GeoDataFrame,
    *,
    max_distance: int | float | None = None,
    fix_double: bool = True,
    ignore_index: bool = False,
    aggfunc: str | dict | list | None = None,
    predicate: str = "intersects",
    grid_size=None,
    n_jobs: int = 1,
    **kwargs,
) -> tuple[GeoDataFrame]:
    """Dissolves selected polygons with the largest neighbor polygon.

    Eliminates selected geometries by dissolving them with the neighboring
    polygon with the largest area. The index and column values of the
    large polygons will be kept, unless else is specified.

    Args:
        gdf: GeoDataFrame with polygon geometries, or a list of GeoDataFrames.
        to_eliminate: The geometries to be eliminated by 'gdf'.
        max_distance: Max distance to search for neighbors. Defaults to None, meaning
            0.
        fix_double: If True, geometries to be eliminated will be erased by overlapping
            geometries to not get double surfaces if the geometries in 'to_eliminate'
            overlaps with multiple geometries in 'gdf'.
        ignore_index: If False (default), the resulting GeoDataFrame will keep the
            index of the large polygons. If True, the resulting axis will be labeled
            0, 1, …, n - 1.
        aggfunc: Aggregation function(s) to use when dissolving/eliminating.
            Defaults to None, meaning the values of 'gdf' is used. Otherwise,
            aggfunc will be passed to pandas groupby.agg. note: The geometries of
            'gdf' are sorted first, but if 'gdf' has missing values, the resulting
            polygons might get values from the polygons to be eliminated
            (if aggfunc="first").
        predicate: Binary predicate passed to sjoin. Defaults to "intersects".
        grid_size: Rounding of the coordinates. Defaults to None.
        n_jobs: Number of threads to use. Defaults to 1.
        **kwargs: Keyword arguments passed to the dissolve method.

    Returns:
        A tuple of the GeoDataFrame with the geometries of 'to_eliminate'
        dissolved in and a GeoDataFrame with the potentionally isolated
        polygons that could not be eliminated. If multiple GeoDataFrame
        are passed as 'gdf', the returned tuple will contain each frame
        plus the isolated polygons as the last item.

    Examples:
    ---------
    Create two polygons with a sliver in between:

    >>> sliver = sg.to_gdf(Polygon([(0, 0), (0.1, 1), (0, 2), (-0.1, 1)]))
    >>> small_poly = sg.to_gdf(
    ...     Polygon([(0, 0), (-0.1, 1), (0, 2), (-1, 2), (-2, 2), (-1, 1)])
    ... )
    >>> large_poly = sg.to_gdf(
    ...     Polygon([(0, 0), (0.1, 1), (1, 2), (2, 2), (3, 2), (3, 0)])
    ... )

    Using multiple GeoDataFrame as input, the sliver is eliminated into
    the large polygon.

    >>> small_poly_eliminated, large_poly_eliminated, isolated = sg.eliminate_by_largest(
    ...     [small_poly, large_poly], sliver
    ... )

    With only one input GeoDataFrame:

    >>> polys = pd.concat([small_poly, large_poly])
    >>> eliminated, isolated = sg.eliminate_by_largest(polys, sliver)
    """
    return _eliminate_by_area(
        gdf,
        to_eliminate=to_eliminate,
        max_distance=max_distance,
        ignore_index=ignore_index,
        sort_ascending=False,
        aggfunc=aggfunc,
        predicate=predicate,
        fix_double=fix_double,
        grid_size=grid_size,
        n_jobs=n_jobs,
        **kwargs,
    )


def eliminate_by_smallest(
    gdf: GeoDataFrame | list[GeoDataFrame],
    to_eliminate: GeoDataFrame,
    *,
    max_distance: int | float | None = None,
    ignore_index: bool = False,
    aggfunc: str | dict | list | None = None,
    predicate: str = "intersects",
    fix_double: bool = True,
    grid_size=None,
    n_jobs: int = 1,
    **kwargs,
) -> tuple[GeoDataFrame]:
    return _eliminate_by_area(
        gdf,
        to_eliminate=to_eliminate,
        max_distance=max_distance,
        ignore_index=ignore_index,
        sort_ascending=True,
        aggfunc=aggfunc,
        predicate=predicate,
        fix_double=fix_double,
        grid_size=grid_size,
        n_jobs=n_jobs,
        **kwargs,
    )


def _eliminate_by_area(
    gdf: GeoDataFrame,
    to_eliminate: GeoDataFrame,
    max_distance: int | float | None,
    sort_ascending: bool,
    ignore_index: bool = False,
    aggfunc: str | dict | list | None = None,
    predicate="intersects",
    fix_double: bool = True,
    grid_size=None,
    n_jobs: int = 1,
    **kwargs,
) -> GeoDataFrame:
    _recurse = kwargs.pop("_recurse", False)

    if not len(to_eliminate) or not len(gdf):
        return gdf, to_eliminate

    if isinstance(gdf, (list, tuple)):
        was_multiple_gdfs = True
        original_cols = [df.columns for df in gdf]
        gdf = pd.concat(df.assign(**{"_df_idx": i}) for i, df in enumerate(gdf))
    else:
        was_multiple_gdfs = False

    crs = gdf.crs
    geom_type = get_geom_type(gdf)

    if not ignore_index:
        idx_mapper = dict(enumerate(gdf.index))
        idx_name = gdf.index.name
        idx_mapper_to_eliminate = dict(enumerate(to_eliminate.index))
        idx_name_to_eliminate = to_eliminate.index.name

    gdf = make_all_singlepart(gdf).reset_index(drop=True)
    to_eliminate = make_all_singlepart(to_eliminate).reset_index(drop=True)

    gdf["_area"] = gdf.area
    gdf["_dissolve_idx"] = gdf.index

    if max_distance:
        to_join = gdf[["_area", "_dissolve_idx", "geometry"]]
        to_join.geometry = to_join.buffer(max_distance)
        joined = to_eliminate.sjoin(to_join, predicate=predicate, how="left")
    else:
        joined = to_eliminate.sjoin(
            gdf[["_area", "_dissolve_idx", "geometry"]], predicate=predicate, how="left"
        )

    # as DataFrames because GeoDataFrame constructor is expensive
    joined = (
        pd.DataFrame(joined)
        .drop(columns="index_right")
        .sort_values("_area", ascending=sort_ascending)
        .loc[lambda x: ~x.index.duplicated(keep="first")]
    )

    gdf = pd.DataFrame(gdf)

    notna = joined.loc[lambda x: x["_dissolve_idx"].notna()]

    eliminated = _eliminate(
        gdf,
        notna,
        aggfunc,
        crs,
        fix_double=fix_double,
        grid_size=grid_size,
        n_jobs=n_jobs,
        **kwargs,
    )

    eliminated = eliminated.drop(
        ["_dissolve_idx", "_area", "_eliminate_idx", "_dissolve_idx"],
        axis=1,
        errors="ignore",
    )

    out = GeoDataFrame(
        eliminated,
        geometry="geometry",
        crs=crs,
    ).pipe(clean_geoms)

    isolated = (
        GeoDataFrame(
            joined.loc[joined["_dissolve_idx"].isna()], geometry="geometry", crs=crs
        )
        .drop(
            ["_dissolve_idx", "_area", "_eliminate_idx", "_dissolve_idx"],
            axis=1,
            errors="ignore",
        )
        .pipe(clean_geoms)
    )

    if not ignore_index:
        out.index = out.index.map(idx_mapper)
        out.index.name = idx_name
        isolated.index = isolated.index.map(idx_mapper_to_eliminate)
        isolated.index.name = idx_name_to_eliminate

    if geom_type != "mixed":
        out = to_single_geom_type(out, geom_type)

    out = out.reset_index(drop=True) if ignore_index else out

    if not _recurse and len(isolated):
        out, isolated = _recursively_eliminate_new_neighbors(
            out,
            isolated,
            func=_eliminate_by_area,
            max_distance=max_distance,
            sort_ascending=sort_ascending,
            fix_double=fix_double,
            predicate=predicate,
            ignore_index=ignore_index,
            aggfunc=aggfunc,
            grid_size=grid_size,
            n_jobs=n_jobs,
        )

    if not was_multiple_gdfs:
        return out, isolated

    for k, v in locals().items():
        try:
            print(k, v.columns)
        except Exception:
            pass

    gdfs = ()
    for i, cols in enumerate(original_cols):
        df = out.loc[out["_df_idx"] == i, cols]
        gdfs += (df,)

    return (*gdfs, isolated)


def _eliminate(
    gdf, to_eliminate, aggfunc, crs, fix_double, grid_size, n_jobs, **kwargs
):
    if not len(to_eliminate):
        return gdf

    gdf["_range_idx_elim"] = range(len(gdf))

    in_to_eliminate = gdf["_dissolve_idx"].isin(to_eliminate["_dissolve_idx"])
    to_dissolve = gdf.loc[in_to_eliminate]
    not_to_dissolve = gdf.loc[~in_to_eliminate].set_index("_dissolve_idx")

    to_eliminate["_to_eliminate"] = 1

    if aggfunc is None:
        concatted = pd.concat(
            [to_dissolve, to_eliminate[["_to_eliminate", "_dissolve_idx", "geometry"]]]
        )
        aggfunc = "first"
    else:
        concatted = pd.concat([to_dissolve, to_eliminate])

    one_hit = concatted.loc[
        lambda x: (x.groupby("_dissolve_idx").transform("size") == 1)
        & (x["_dissolve_idx"].notna())
    ].set_index("_dissolve_idx")

    assert len(one_hit) == 0

    many_hits = concatted.loc[
        lambda x: x.groupby("_dissolve_idx").transform("size") > 1
    ]

    if not len(many_hits):
        return one_hit

    # aggregate all columns except geometry
    kwargs.pop("as_index", None)
    eliminated = (
        many_hits.drop(columns="geometry")
        .groupby("_dissolve_idx", **kwargs)
        .agg(aggfunc)
        .drop(["_area"], axis=1, errors="ignore")
    )

    # aggregate geometry
    if fix_double:
        assert eliminated.index.is_unique

        many_hits = many_hits.set_index("_dissolve_idx")
        many_hits["_row_idx"] = range(len(many_hits))

        # TODO kan dette fikses trygt med .duplicated og ~x.duplicated?
        eliminators: pd.Series = many_hits.loc[
            many_hits["_to_eliminate"] != 1, "geometry"
        ]
        to_be_eliminated = many_hits.loc[many_hits["_to_eliminate"] == 1]

        # all_geoms: pd.Series = gdf.set_index("_dissolve_idx").geometry
        all_geoms: pd.Series = gdf.geometry

        # more_than_one = get_num_geometries(all_geoms.values) > 1
        # all_geoms.loc[more_than_one] = all_geoms.loc[more_than_one].apply(
        #     _unary_union_for_notna
        # )

        # more_than_one = get_num_geometries(to_be_eliminated.values) > 1
        # to_be_eliminated.loc[more_than_one, "geometry"] = to_be_eliminated.loc[
        #     more_than_one, "geometry"
        # ].apply(_unary_union_for_notna)

        # create DataFrame of intersection pairs
        tree = STRtree(all_geoms.values)
        left, right = tree.query(
            to_be_eliminated.geometry.values, predicate="intersects"
        )

        pairs = pd.Series(right, index=left).to_frame("right")
        pairs["_dissolve_idx"] = pairs.index.map(
            dict(enumerate(to_be_eliminated.index))
        )

        # pairs = pairs.loc[lambda x: x["right"] != x["_dissolve_idx"]]

        soon_erased = to_be_eliminated.iloc[pairs.index]
        intersecting = all_geoms.iloc[pairs["right"]]

        shoud_not_erase = soon_erased.index != intersecting.index
        soon_erased = soon_erased[shoud_not_erase]
        intersecting = intersecting[shoud_not_erase]

        missing = to_be_eliminated.loc[
            # (~to_be_eliminated.index.isin(soon_erased.index))
            # |
            (~to_be_eliminated["_row_idx"].isin(soon_erased["_row_idx"])),
            # | (~to_be_eliminated["_row_idx"].isin(soon_erased.index)),
            "geometry",
        ]

        # allign and aggregate by dissolve index to not get duplicates in difference
        intersecting.index = soon_erased.index

        soon_erased = _grouped_unary_union(soon_erased, level=0, grid_size=grid_size)
        intersecting = _grouped_unary_union(intersecting, level=0, grid_size=grid_size)

        assert soon_erased.index.equals(soon_erased.index)

        # soon_erased = soon_erased.geometry.groupby(level=0).agg(
        #     lambda x: unary_union(x, grid_size=grid_size)
        # )
        # intersecting = intersecting.groupby(level=0).agg(
        #     lambda x: unary_union(x, grid_size=grid_size)
        # )

        # explore_locals(center=_DEBUG_CONFIG["center"])

        soon_erased.loc[:] = _try_difference(
            soon_erased.to_numpy(),
            intersecting.to_numpy(),
            grid_size=grid_size,
            n_jobs=n_jobs,
            geom_type="polygon",
        )

        missing = _grouped_unary_union(missing, level=0, grid_size=grid_size)

        missing = make_all_singlepart(missing).loc[lambda x: x.area > 0]

        soon_erased = make_all_singlepart(soon_erased).loc[lambda x: x.area > 0]

        if 0:
            tree = STRtree(soon_erased.values)
            left, right = tree.query(missing.values, predicate="intersects")
            explore_locals(
                missing2=to_gdf(missing.to_numpy()[left], 25833),
                soon_erased2=to_gdf(soon_erased.to_numpy()[right], 25833),
                center=_DEBUG_CONFIG["center"],
            )
            missing = pd.Series(
                difference(
                    missing.to_numpy()[left],
                    soon_erased.to_numpy()[right],
                    grid_size=grid_size,
                ),
                index=left,
            ).loc[lambda x: (x.notna()) & (~is_empty(x))]

        soon_eliminated = pd.concat([eliminators, soon_erased, missing])
        more_than_one = get_num_geometries(soon_eliminated.values) > 1

        soon_eliminated.loc[more_than_one] = soon_eliminated.loc[more_than_one].apply(
            _unary_union_for_notna
        )

        if n_jobs > 1:
            eliminated["geometry"] = GeoSeries(
                _parallel_unary_union_geoseries(
                    soon_eliminated,
                    level=0,
                    grid_size=grid_size,
                    n_jobs=n_jobs,
                ),
                index=eliminated.index,
            )
        else:
            eliminated["geometry"] = _grouped_unary_union(soon_eliminated, level=0)
            # eliminated["geometry"] = soon_eliminated.groupby(level=0).agg(
            #     lambda x: make_valid(unary_union(x))
            # )

    else:
        if n_jobs > 1:
            eliminated["geometry"] = _parallel_unary_union(
                many_hits, by="_dissolve_idx", grid_size=grid_size, n_jobs=n_jobs
            )
        else:
            eliminated["geometry"] = _grouped_unary_union(many_hits, by="_dissolve_idx")

    # setting crs on the GeometryArrays to avoid warning in concat
    not_to_dissolve.geometry.values.crs = crs
    try:
        eliminated.geometry.values.crs = crs
    except AttributeError:
        pass
    one_hit.geometry.values.crs = crs

    to_concat = [not_to_dissolve, eliminated, one_hit]

    assert all(df.index.name == "_dissolve_idx" for df in to_concat)

    out = pd.concat(to_concat).sort_index()

    duplicated_geoms = _get_intersecting_geometries(
        GeoDataFrame(
            {
                "geometry": out.geometry.values,
                "_range_idx_elim_dups": out["_range_idx_elim"].values,
            },
        ),
        geom_type="polygon",
        keep_geom_type=True,
        n_jobs=n_jobs,
        predicate="intersects",
    ).pipe(clean_geoms)
    duplicated_geoms.geometry = duplicated_geoms.buffer(-PRECISION)
    duplicated_geoms = duplicated_geoms.pipe(clean_geoms)

    if len(duplicated_geoms):
        hits_in_original_df = duplicated_geoms.sjoin(
            GeoDataFrame(
                {
                    "geometry": gdf.geometry.values,
                    "_range_idx_elim": gdf["_range_idx_elim"].values,
                },
            ),
            how="inner",
        )

        should_be_erased = hits_in_original_df.loc[
            lambda x: x["_range_idx_elim"] != x["_range_idx_elim_dups"]
        ]

        should_be_erased_idx = list(
            sorted(should_be_erased["_range_idx_elim_dups"].unique())
        )
        should_erase = (
            should_be_erased.groupby("_range_idx_elim_dups")["geometry"]
            .agg(lambda x: make_valid(union_all(x)))
            .sort_index()
        )

        # aligining out with "should_erase" before rowwise difference
        out = out.sort_values("_range_idx_elim")
        assert out["_range_idx_elim"].is_unique
        to_be_erased_idx = out["_range_idx_elim"].isin(should_be_erased_idx)

        out.loc[to_be_erased_idx, "geometry"] = make_valid(
            difference(
                out.loc[
                    to_be_erased_idx,
                    "geometry",
                ].values,
                should_erase.values,
            )
        )

        from ..maps.maps import explore

        # display(hits_in_original_df)
        # display(should_be_erased.assign(area=lambda x: x.area))

        explore(
            gdf=to_gdf(gdf, 25833),
            out=to_gdf(out, 25833),
            should_be_erased=to_gdf(should_be_erased, 25833),
            duplicated_geoms=duplicated_geoms.set_crs(25833),
            eli=GeoDataFrame(
                {
                    "geometry": out.geometry.values,
                    "_range_idx_elim": out["_range_idx_elim"].values,
                },
                crs=25833,
            ),
            center=_DEBUG_CONFIG["center"],
        )

    _try_debug_print("inni _eliminate")
    _try_debug_print(duplicated_geoms)
    explore_locals(center=_DEBUG_CONFIG["center"])

    return out.drop(columns=["_to_eliminate", "_range_idx_elim"])


def clean_dissexp(df: GeoDataFrame, dissolve_func: Callable, **kwargs) -> GeoDataFrame:
    """Experimental."""
    original_points = GeoDataFrame(
        {"geometry": get_parts(extract_unique_points(df.geometry.values))}
    )[lambda x: ~x.geometry.duplicated()]

    dissolved = df.copy()

    try:
        dissolved.geometry = dissolved.buffer(PRECISION, resolution=1, join_style=2)
    except AttributeError as e:
        if isinstance(dissolved, GeoSeries):
            dissolved.loc[:] = dissolved.buffer(PRECISION, resolution=1, join_style=2)
        else:
            raise e

    dissolved = dissolve_func(dissolved, **kwargs)

    try:
        dissolved.geometry = dissolved.buffer(-PRECISION, resolution=1, join_style=2)
    except AttributeError as e:
        if isinstance(dissolved, GeoSeries):
            dissolved.loc[:] = dissolved.buffer(-PRECISION, resolution=1, join_style=2)
        else:
            raise e

    dissolved = dissolved.loc[lambda x: ~x.geometry.is_empty]
    dissolved = dissolved.explode(ignore_index=True)

    original_points = sfilter_inverse(original_points, dissolved.buffer(-PRECISION))

    snapped = (
        PolygonsAsRings(
            dissolved.geometry,
        )
        .apply_numpy_func(
            _snap_points_back,
            kwargs={"snap_to": original_points, "tolerance": PRECISION},
        )
        .to_numpy()
    )

    try:
        dissolved.geometry = snapped
    except AttributeError as e:
        if isinstance(dissolved, GeoSeries):
            dissolved.loc[:] = snapped
        else:
            raise e

    return dissolved


def _snap_points_back(rings, snap_to, tolerance):
    points = GeoDataFrame({"geometry": extract_unique_points(rings)})
    points = points.explode(index_parts=True)

    snap_to["geom_right"] = snap_to.geometry
    nearest = points.sjoin_nearest(snap_to, max_distance=tolerance)
    points.loc[nearest.index, points.geometry.name] = nearest["geom_right"]

    new_rings = points.groupby(level=0)[points.geometry.name].agg(LinearRing)
    return new_rings


def close_thin_holes(gdf: GeoDataFrame, tolerance: int | float) -> GeoDataFrame:
    gdf = make_all_singlepart(gdf)
    holes = get_holes(gdf)
    inside_holes = union_all(sfilter(gdf, holes, predicate="within").geometry.values)

    def to_none_if_thin(geoms):
        if not len(geoms):
            return geoms
        try:
            polys = polygons(geoms)
        except GEOSException:
            polys = make_valid(polygons(make_valid(geoms)))
        if inside_holes is not None:
            polys = difference(polys, inside_holes)
        buffered_in = buffer(polys, -(tolerance / 2))
        return np.where(is_empty(buffered_in), None, geoms)

    if not (gdf.geom_type == "Polygon").all():
        raise ValueError(gdf.geom_type.value_counts())

    return PolygonsAsRings(gdf).apply_numpy_func_to_interiors(to_none_if_thin).to_gdf()


def close_all_holes(
    gdf: GeoDataFrame | GeoSeries,
    *,
    ignore_islands: bool = False,
    copy: bool = True,
) -> GeoDataFrame | GeoSeries:
    """Closes all holes in polygons.

    It takes a GeoDataFrame or GeoSeries of polygons and
    returns the outer circle.

    Args:
        gdf: GeoDataFrame or GeoSeries of polygons.
        copy: if True (default), the input GeoDataFrame or GeoSeries is copied.
            Defaults to True.
        ignore_islands: If False (default), polygons inside the holes (islands)
            will be erased from the output geometries. If True, the entire
            holes will be closed and the islands kept, meaning there might be
            duplicate surfaces in the resulting geometries.
            Note that ignoring islands is a lot faster.

    Returns:
        A GeoDataFrame or GeoSeries of polygons with closed holes in the geometry
        column.

    Examples:
    ---------
    Let's create a circle with a hole in it.

    >>> point = sg.to_gdf([260000, 6650000], crs=25833)
    >>> point
                            geometry
    0  POINT (260000.000 6650000.000)
    >>> circle = sg.buff(point, 1000)
    >>> small_circle = sg.buff(point, 500)
    >>> circle_with_hole = circle.overlay(small_circle, how="difference")
    >>> circle_with_hole.area
    0    2.355807e+06
    dtype: float64

    Close the hole.

    >>> holes_closed = sg.close_all_holes(circle_with_hole)
    >>> holes_closed.area
    0    3.141076e+06
    dtype: float64
    """
    if not isinstance(gdf, (GeoDataFrame, GeoSeries)):
        raise ValueError(
            f"'gdf' should be of type GeoDataFrame or GeoSeries. Got {type(gdf)}"
        )

    if not len(gdf):
        return gdf

    if copy:
        gdf = gdf.copy()

    gdf = make_all_singlepart(gdf)

    if ignore_islands:
        geoms = gdf.geometry if isinstance(gdf, GeoDataFrame) else gdf
        holes_closed = make_valid(polygons(get_exterior_ring(geoms)))
        if isinstance(gdf, GeoDataFrame):
            gdf.geometry = holes_closed
            return gdf
        elif isinstance(gdf, GeoSeries):
            return GeoSeries(holes_closed, crs=gdf.crs)
        else:
            return holes_closed

    all_geoms = make_valid(union_all(gdf.geometry.values))
    if isinstance(gdf, GeoDataFrame):
        gdf.geometry = gdf.geometry.map(
            lambda x: _close_all_holes_no_islands(x, all_geoms)
        )
        return gdf
    else:
        return gdf.map(lambda x: _close_all_holes_no_islands(x, all_geoms))


def close_small_holes(
    gdf: GeoDataFrame | GeoSeries,
    max_area: int | float,
    *,
    ignore_islands: bool = False,
    copy: bool = True,
) -> GeoDataFrame | GeoSeries:
    """Closes holes in polygons if the area is less than the given maximum.

    It takes a GeoDataFrame or GeoSeries of polygons and
    fills the holes that are smaller than the specified area given in units of
    either square meters ('max_m2') or square kilometers ('max_km2').

    Args:
        gdf: GeoDataFrame or GeoSeries of polygons.
        max_area: The maximum area in the unit of the GeoDataFrame's crs.
        ignore_islands: If False (default), polygons inside the holes (islands)
            will be erased from the "hole" geometries before the area is calculated.
            If True, the entire polygon interiors will be considered, meaning there
            might be duplicate surfaces in the resulting geometries.
            Note that ignoring islands is a lot faster.
        copy: if True (default), the input GeoDataFrame or GeoSeries is copied.
            Defaults to True.

    Returns:
        A GeoDataFrame or GeoSeries of polygons with closed holes in the geometry
        column.

    Raises:
        ValueError: If the coordinate reference system of the GeoDataFrame is not in
            meter units.
        ValueError: If both 'max_m2' and 'max_km2' is given.

    Examples:
    ---------
    Let's create a circle with a hole in it.

    >>> point = sg.to_gdf([260000, 6650000], crs=25833)
    >>> point
                            geometry
    0  POINT (260000.000 6650000.000)
    >>> circle = sg.buff(point, 1000)
    >>> small_circle = sg.buff(point, 500)
    >>> circle_with_hole = circle.overlay(small_circle, how="difference")
    >>> circle_with_hole.area
    0    2.355807e+06
    dtype: float64

    Close holes smaller than 1 square kilometer (1 million square meters).

    >>> holes_closed = sg.close_small_holes(circle_with_hole, max_area=1_000_000)
    >>> holes_closed.area
    0    3.141076e+06
    dtype: float64

    The hole will not be closed if it is larger.

    >>> holes_closed = sg.close_small_holes(circle_with_hole, max_area=1_000)
    >>> holes_closed.area
    0    2.355807e+06
    dtype: float64
    """
    if not isinstance(gdf, (GeoSeries, GeoDataFrame)):
        raise ValueError(
            f"'gdf' should be of type GeoDataFrame or GeoSeries. Got {type(gdf)}"
        )

    if not len(gdf):
        return gdf

    if copy:
        gdf = gdf.copy()

    gdf = make_all_singlepart(gdf)

    if not ignore_islands:
        all_geoms = make_valid(union_all(gdf.geometry.values))

        if isinstance(gdf, GeoDataFrame):
            gdf.geometry = gdf.geometry.map(
                lambda x: _close_small_holes_no_islands(x, max_area, all_geoms)
            )
            return gdf
        else:
            return gdf.map(
                lambda x: _close_small_holes_no_islands(x, max_area, all_geoms)
            )
    else:
        geoms = (
            gdf.geometry.to_numpy() if isinstance(gdf, GeoDataFrame) else gdf.to_numpy()
        )
        exteriors = get_exterior_ring(geoms)
        assert len(exteriors) == len(geoms)

        max_rings = max(get_num_interior_rings(geoms))

        if not max_rings:
            return gdf

        # looping through max for all geoms since arrays must be equal length
        interiors = np.array(
            [[get_interior_ring(geom, i) for i in range(max_rings)] for geom in geoms]
        )
        assert interiors.shape == (len(geoms), max_rings), interiors.shape

        areas = area(polygons(interiors))
        interiors[(areas < max_area) | np.isnan(areas)] = None

        results = polygons(exteriors, interiors)

        if isinstance(gdf, GeoDataFrame):
            gdf.geometry = results
            return gdf
        else:
            return GeoSeries(results, crs=gdf.crs)


def _close_small_holes_no_islands(poly, max_area, all_geoms):
    """Closes small holes within one shapely geometry of polygons."""
    # start with a list containing the polygon,
    # then append all holes smaller than 'max_km2' to the list.
    holes_closed = [poly]
    singlepart = get_parts(poly)
    for part in singlepart:
        n_interior_rings = get_num_interior_rings(part)

        if not (n_interior_rings):
            continue

        for n in range(n_interior_rings):
            hole = polygons(get_interior_ring(part, n))
            try:
                no_islands = union_all(hole.difference(all_geoms))
            except GEOSException:
                no_islands = make_valid(union_all(hole.difference(all_geoms)))

            if area(no_islands) < max_area:
                holes_closed.append(no_islands)

    return make_valid(union_all(holes_closed))


def _close_all_holes_no_islands(poly, all_geoms):
    """Closes all holes within one shapely geometry of polygons."""
    # start with a list containing the polygon,
    # then append all holes smaller than 'max_km2' to the list.
    holes_closed = [poly]
    singlepart = get_parts(poly)
    for part in singlepart:
        n_interior_rings = get_num_interior_rings(part)

        for n in range(n_interior_rings):
            hole = polygons(get_interior_ring(part, n))
            try:
                no_islands = union_all(hole.difference(all_geoms))
            except GEOSException:
                no_islands = make_valid(union_all(hole.difference(all_geoms)))

            holes_closed.append(no_islands)

    return make_valid(union_all(holes_closed))


def get_gaps(
    gdf: GeoDataFrame,
    include_interiors: bool = False,
    grid_size: float | int | None = None,
) -> GeoDataFrame:
    """Get the gaps between polygons.

    Args:
        gdf: GeoDataFrame of polygons.
        include_interiors: If False (default), the holes inside individual polygons
            will not be included as gaps.
        grid_size: Rounding of the coordinates.

    Note:
        See get_holes to find holes inside singlepart polygons.

    Returns:
        GeoDataFrame of polygons with only a geometry column.
    """
    if not len(gdf):
        return GeoDataFrame({"geometry": []}, crs=gdf.crs)

    if not include_interiors:
        gdf = close_all_holes(gdf)

    bbox = GeoDataFrame(
        {"geometry": [box(*tuple(gdf.total_bounds)).buffer(1)]}, crs=gdf.crs
    )

    bbox_diff = make_all_singlepart(
        clean_overlay(
            bbox, gdf, how="difference", geom_type="polygon", grid_size=grid_size
        )
    )

    # remove the outer "gap", i.e. the surrounding area
    bbox_ring = get_exterior_ring(bbox.geometry.values)
    without_outer_ring = sfilter_inverse(bbox_diff, bbox_ring)
    return without_outer_ring.reset_index(drop=True)


def get_holes(gdf: GeoDataFrame, as_polygons: bool = True) -> GeoDataFrame:
    """Get the holes inside polygons.

    Args:
        gdf: GeoDataFrame of polygons.
        as_polygons: If True (default), the holes will be returned as polygons.
            If False, they will be returned as LinearRings.

    Note:
        See get_gaps to find holes/gaps between undissolved polygons.

    Returns:
        GeoDataFrame of polygons or linearrings with only a geometry column.
    """
    if not len(gdf):
        return GeoDataFrame({"geometry": []}, index=gdf.index, crs=gdf.crs)

    def as_linearring(x):
        return x

    astype = polygons if as_polygons else as_linearring

    geoms = make_all_singlepart(gdf.geometry).to_numpy()

    rings = [
        GeoSeries(astype(get_interior_ring(geoms, i)), crs=gdf.crs)
        for i in range(max(get_num_interior_rings(geoms)))
    ]

    return (
        GeoDataFrame({"geometry": (pd.concat(rings).pipe(clean_geoms).sort_index())})
        if rings
        else GeoDataFrame({"geometry": []}, crs=gdf.crs)
    )


def split_polygons_by_lines(polygons: GeoSeries, lines: GeoSeries) -> GeoSeries:
    idx_mapper = dict(enumerate(polygons.index))
    idx_name = polygons.index.name
    polygons = polygons.copy()
    polygons.index = range(len(polygons))

    # use pandas to explode faster (from list instead of GeoSeries.explode)
    splitted = pd.Series(polygons.geometry.to_numpy())
    lines = to_geoseries(lines)
    lines.index = range(len(lines))

    # find intersection pairs to split relevant polygon for each line
    tree = STRtree(splitted.values)
    left, right = tree.query(lines.values, predicate="intersects")
    pairs = pd.Series(right, index=left)

    lines = lines.loc[lambda x: x.index.isin(pairs.index)]

    for i, line in lines.items():
        intersecting = pairs.loc[[i]].values
        try:
            splitted.loc[intersecting] = splitted.loc[intersecting].apply(
                lambda poly: SplitOp._split_polygon_with_line(poly, line) or poly
            )
        except TypeError:
            # if we got multipolygon
            splitted = splitted.apply(get_parts).explode()
            splitted.loc[intersecting] = splitted.loc[intersecting].apply(
                lambda poly: SplitOp._split_polygon_with_line(poly, line) or poly
            )
        splitted = splitted.explode()

    if isinstance(polygons, GeoDataFrame):
        polygons = polygons.loc[splitted.index]
        polygons.geometry = splitted
        polygons.index = polygons.index.map(idx_mapper)
        polygons.index.name = idx_name
        return polygons
    else:
        splitted.index = splitted.index.map(idx_mapper)
        splitted.index.name = idx_name
        return splitted


def split_by_neighbors(
    df: GeoDataFrame,
    split_by: GeoDataFrame,
    tolerance: int | float,
    grid_size: float | int | None = None,
) -> GeoDataFrame:
    if not len(df):
        return df

    df = make_all_singlepart(df)

    split_by = split_by.copy()

    intersecting_lines = (
        clean_overlay(
            to_lines(split_by),
            df.buffer(tolerance).to_frame("geometry"),
            how="intersection",
            grid_size=grid_size,
        )
        .pipe(get_line_segments)
        .reset_index(drop=True)
    )

    endpoints = intersecting_lines.boundary.explode(index_parts=False)

    lines = extend_lines(
        endpoints.loc[lambda x: ~x.index.duplicated(keep="first")].values,
        endpoints.loc[lambda x: ~x.index.duplicated(keep="last")].values,
        distance=tolerance * 3,
    )

    return split_polygons_by_lines(df, lines)



"""Interactive map of one or more GeoDataFrames with layers that can be toggles on/off.

This module holds the Explore class, which is the basis for the explore, samplemap and
clipmap functions from the 'maps' module.
"""

import os
import random
import re
import warnings
from collections.abc import Iterable
from numbers import Number
from pathlib import Path
from statistics import mean
from typing import Any
from typing import ClassVar

import branca as bc
import folium
import geopandas as gpd
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import shapely
import xyzservices
from folium import plugins
from geopandas import GeoDataFrame
from geopandas import GeoSeries
from IPython.display import display
from jinja2 import Template
from pandas.api.types import is_datetime64_any_dtype
from shapely import Geometry
from shapely import box
from shapely import union_all
from shapely.geometry import LineString

from ..geopandas_tools.bounds import get_total_bounds
from ..geopandas_tools.conversion import to_bbox
from ..geopandas_tools.conversion import to_gdf
from ..geopandas_tools.conversion import to_shapely
from ..geopandas_tools.general import clean_geoms
from ..geopandas_tools.general import make_all_singlepart
from ..geopandas_tools.geometry_types import get_geom_type
from ..geopandas_tools.geometry_types import to_single_geom_type

try:
    from ..raster.image_collection import Band
    from ..raster.image_collection import Image
    from ..raster.image_collection import ImageCollection
except ImportError:

    class Band:
        """Placeholder."""

    class Image:
        """Placeholder."""

    class ImageCollection:
        """Placeholder."""


from .httpserver import run_html_server
from .map import Map
from .map import _determine_best_name
from .tilesources import kartverket
from .tilesources import xyz

try:
    from torchgeo.datasets.geo import RasterDataset
except ImportError:

    class RasterDataset:
        """Placeholder."""


# the geopandas._explore raises a deprication warning. Ignoring for now.
warnings.filterwarnings(
    action="ignore", category=matplotlib.MatplotlibDeprecationWarning
)
pd.options.mode.chained_assignment = None


# gray for NaNs
NAN_COLOR = "#969696"


# cols to not show when hovering over geometries (tooltip)
COLS_TO_DROP = ["color", "col_as_int", "geometry"]


# from geopandas
_MAP_KWARGS = [
    "location",
    "prefer_canvas",
    "no_touch",
    "disable_3d",
    "png_enabled",
    "zoom_control",
    "crs",
    "zoom_start",
    "left",
    "top",
    "position",
    "min_zoom",
    "max_zoom",
    "min_lat",
    "max_lat",
    "min_lon",
    "max_lon",
    "max_bounds",
]


class MeasureControlFix(plugins.MeasureControl):
    """Monkey-patch to fix a bug in the lenght measurement control.

    Kudos to abewartech (https://github.com/ljagis/leaflet-measure/issues/171).
    """

    _template = Template(
        """
{% macro script(this, kwargs) %}
    L.Control.Measure.include({ _setCaptureMarkerIcon: function () { this._captureMarker.options.autoPanOnFocus = false; this._captureMarker.setIcon( L.divIcon({ iconSize: this._map.getSize().multiplyBy(2), }), ); }, });
    var {{ this.get_name() }} = new L.Control.Measure(
        {{ this.options|tojson }});
    {{this._parent.get_name()}}.addControl({{this.get_name()}});

{% endmacro %}
    """
    )

    def __init__(
        self, active_color: str = "red", completed_color: str = "red", **kwargs
    ) -> None:
        """Run super __init__ after the new _template class attribute is made."""
        super().__init__(
            active_color=active_color, completed_color=completed_color, **kwargs
        )


def to_tile(tile: str | xyzservices.TileProvider, max_zoom: int) -> folium.TileLayer:
    common_bgmaps = {
        "openstreetmap": folium.TileLayer(
            "OpenStreetMap", min_zoom=0, max_zoom=max_zoom
        ),
        "grunnkart": kartverket.norges_grunnkart,
        "gråtone": kartverket.norges_grunnkart_gråtone,
        "norge_i_bilder": kartverket.norge_i_bilder,
        "dark": xyz.CartoDB.DarkMatter,
        "voyager": xyz.CartoDB.Voyager,
        "strava": xyz.Strava.All,
    }
    try:
        name = tile["name"]
    except TypeError:
        name = tile

    if not isinstance(tile, str):
        try:
            return folium.TileLayer(tile, name=name, max_zoom=max_zoom)
        except TypeError:
            return folium.TileLayer(tile, max_zoom=max_zoom)

    try:
        provider = common_bgmaps[tile.lower()]
    except KeyError:
        provider = xyzservices.providers.query_name(tile)

    if isinstance(provider, folium.TileLayer):
        return provider

    if isinstance(provider, xyzservices.TileProvider):
        attr = provider.html_attribution
        provider = provider.build_url(scale_factor="{r}")
    else:
        try:
            attr = provider["attr"]
        except (AttributeError, TypeError):
            attr = None

    return folium.TileLayer(provider, name=name, attr=attr, max_zoom=max_zoom)


def _single_band_to_arr(band, mask, name, raster_data_dict):
    try:
        arr = band.values
    except (ValueError, AttributeError):
        arr = band.load(indexes=1, bounds=mask).values
    bounds: tuple = (
        _any_to_bbox_crs4326(mask, band.crs)
        if mask is not None
        else union_all(
            gpd.GeoSeries(box(*band.bounds), crs=band.crs).to_crs(4326).geometry.values
        ).bounds
    )
    try:
        raster_data_dict["cmap"] = band.get_cmap(arr)
    except Exception:
        try:
            raster_data_dict["cmap"] = plt.get_cmap(band.cmap)
        except Exception:
            raster_data_dict["cmap"] = band.cmap or "Grays"
    raster_data_dict["arr"] = arr
    raster_data_dict["bounds"] = bounds
    raster_data_dict["label"] = name


def _any_to_bbox_crs4326(obj, crs):
    return to_bbox(to_gdf(obj, crs).to_crs(4326))


class Explore(Map):
    """Class for displaying and saving html maps of multiple GeoDataFrames."""

    # class attribute that can be overridden locally
    tiles: ClassVar[tuple[str]] = (
        "OpenStreetMap",
        "dark",
        "norge_i_bilder",
        # "grunnkart",
    )

    def __init__(
        self,
        *gdfs,
        mask=None,
        column: str | None = None,
        popup: bool = True,
        max_zoom: int = 40,
        smooth_factor: float = 1.1,
        browser: bool = False,
        prefer_canvas: bool = True,
        measure_control: bool = True,
        geocoder: bool = False,
        out_path: str | None = None,
        show: bool | Iterable[bool] | None = None,
        text: str | None = None,
        decimals: int = 6,
        max_images: int = 10,
        **kwargs,
    ) -> None:
        """Initialiser.

        Args:
            *gdfs: One or more GeoDataFrames.
            mask: Optional mask to clip to.
            column: Optional column to color the data by.
            popup: Whether to make the data popups clickable.
            max_zoom: Max levels of zoom.
            smooth_factor: Float of 1 or higher, 1 meaning no smoothing
                of lines.
            browser: Whether to open the map in a browser tab.
            prefer_canvas: Option.
            measure_control: Whether to include measurement box.
            geocoder: Whether to include search bar for addresses.
            out_path: Optional file path to an html file. The map will then
                be saved instead of displayed.
            show: Whether to show or hide the data upon creating the map.
                If False, the data can be toggled on later. 'show' can also be
                a sequence of boolean values the same length as the number of
                GeoDataFrames.
            max_images: Maximum number of images (Image, ImageCollection, Band) to show per
                map. Defaults to 15.
            text: Optional text for a text box in the map.
            decimals: Number of decimals in the coordinates.
            **kwargs: Additional keyword arguments. Can also be geometry-like objects
                where the key is the label.
        """
        self.popup = popup
        self.max_zoom = max_zoom
        self.smooth_factor = smooth_factor
        self.prefer_canvas = prefer_canvas
        self.measure_control = measure_control
        self.geocoder = geocoder
        self.out_path = out_path
        self.mask = mask
        self.text = text
        self.decimals = decimals
        self.max_images = max_images
        self.legend = None

        self.browser = browser
        if not self.browser and "show_in_browser" in kwargs:
            self.browser = kwargs.pop("show_in_browser")
        if not self.browser and "in_browser" in kwargs:
            self.browser = kwargs.pop("in_browser")

        if show is None:
            self._show_was_none = True
            show = True
        else:
            self._show_was_none = False

        new_gdfs = {}
        self.rasters = {}
        for i, gdf in enumerate(gdfs):
            name = _determine_best_name(gdf, column, i)

            if name in new_gdfs or name in self.rasters:
                name += str(i)

            if isinstance(gdf, (ImageCollection | Image | Band)):
                self.rasters[name] = gdf.copy()
                continue
            try:
                new_gdfs[name] = to_gdf(gdf)
            except Exception:
                continue

        new_kwargs = {}
        for key, value in kwargs.items():
            if isinstance(value, (ImageCollection | Image | Band)):
                self.rasters[key] = value
            else:
                new_kwargs[key] = value

        super().__init__(column=column, show=show, **(new_kwargs | new_gdfs))

        if self.gdfs is None:
            return

        # stringify or remove columns not renerable by leaflet (list, geometry etc.)
        new_gdfs, show_new = [], []
        for gdf, show in zip(self.gdfs, self.show, strict=True):
            try:
                gdf = gdf.reset_index()
            except Exception:
                pass
            for col in gdf.columns:
                if is_datetime64_any_dtype(gdf[col]):
                    try:
                        gdf[col] = [str(x) for x in gdf[col].dt.round("d")]
                    except Exception:
                        gdf = gdf.drop(col, axis=1)
                    continue

                if not len(gdf.loc[gdf[col].notna()]):
                    continue
                if not isinstance(
                    gdf.loc[gdf[col].notna(), col].iloc[0], (Number, str, Geometry)
                ) or (
                    col != gdf._geometry_column_name
                    and isinstance(gdf.loc[gdf[col].notna(), col].iloc[0], (Geometry))
                ):
                    try:
                        gdf[col] = gdf[col].astype(str).fillna(pd.NA)
                    except Exception:
                        gdf = gdf.drop(col, axis=1)

            try:
                gdf.index = gdf.index.astype(str)
            except Exception:
                pass
            new_gdfs.append(to_gdf(gdf))
            show_new.append(show)
        self._gdfs = new_gdfs
        if self._gdfs:
            self._gdf = pd.concat(new_gdfs, ignore_index=True)
        else:
            self._gdf = GeoDataFrame({"geometry": [], self._column: []})
        self.show = show_new

        # if self._show_was_none and len(self._gdfs) > 6:
        #     self.show = [False] * len(self._gdfs)

        if self._is_categorical:
            if len(self.gdfs) == 1:
                self._split_categories()
        else:
            if not self._cmap:
                self._cmap = "viridis"
            self.cmap_start = self.kwargs.pop("cmap_start", 0)
            self.cmap_stop = self.kwargs.pop("cmap_stop", 256)

        # if self._gdf.crs is None:
        #     self.kwargs["crs"] = "Simple"

        self.original_crs = self.gdf.crs

    def __repr__(self) -> str:
        """Representation."""
        return f"{self.__class__.__name__}({len(self)})"

    def __len__(self) -> int:
        """Number of gdfs that have rows plus number of raster images."""
        try:
            rasters = self.raster_data
        except AttributeError:
            rasters = self.rasters
        return len([gdf for gdf in self._gdfs if len(gdf)]) + len(rasters)

    def __bool__(self) -> bool:
        """True if any gdfs have rows or there are any raster images."""
        return bool(len(self))

    def explore(
        self,
        column: str | None = None,
        center: Any | None = None,
        size: int | None = None,
        mask: Any | None = None,
        **kwargs,
    ) -> None:
        """Explore all the data."""
        self.mask = mask if mask is not None else self.mask
        if (
            self._gdfs
            and not any(len(gdf) for gdf in self._gdfs)
            and not len(self.rasters)
        ):
            warnings.warn("None of the GeoDataFrames have rows.", stacklevel=1)
            return
        if column:
            self._column = column
            self._update_column()
            kwargs.pop("column", None)

        if self.mask is not None:
            return self.clipmap(mask=self.mask, column=self._column, **kwargs)

        if center is None:
            self.to_show = self._gdfs
            self._explore(**kwargs)
            return

        size = size if size else 1000

        centerpoint = (
            to_gdf(center, crs=self.crs)
            if not isinstance(center, GeoDataFrame)
            else center
        )

        gdfs: tuple[GeoDataFrame] = ()
        for gdf in self._gdfs:
            keep_geom_type = False if get_geom_type(gdf) == "mixed" else True
            gdf = gdf.clip(centerpoint.buffer(size), keep_geom_type=keep_geom_type)
            gdfs = gdfs + (gdf,)
        self._gdfs = gdfs
        self._gdf = pd.concat(gdfs, ignore_index=True)

        self._get_unique_values()

        self._explore(**kwargs)

    def samplemap(
        self,
        size: int,
        sample: Any,
        column: str | None = None,
        **kwargs,
    ) -> None:
        """Explore a sample of the data."""
        if column:
            self._column = column
            self._update_column()
            kwargs.pop("column", None)

        try:
            sample = sample.sample(1)
        except Exception:
            pass

        try:
            sample = to_gdf(to_shapely(sample)).pipe(make_all_singlepart)
        except Exception:
            sample = to_gdf(to_shapely(to_bbox(sample))).pipe(make_all_singlepart)

        random_point = sample.sample_points(size=1)

        self.center = (random_point.geometry.iloc[0].x, random_point.geometry.iloc[0].y)
        print(f"center={self.center}, size={size}")

        mask = random_point.buffer(size)

        return self.clipmap(mask, column, **kwargs)

    def clipmap(
        self,
        mask: Any,
        column: str | None = None,
        **kwargs,
    ) -> None:
        """Explore the data within a mask extent."""
        self.mask = mask
        if column:
            self._column = column
            self._update_column()
            kwargs.pop("column", None)

        gdfs: tuple[GeoDataFrame] = ()
        for gdf in self._gdfs:
            gdf = gdf.clip(self.mask)
            collections = gdf.loc[gdf.geom_type == "GeometryCollection"]
            if len(collections):
                collections = make_all_singlepart(collections)
                gdf = pd.concat([gdf, collections], ignore_index=False)
            gdfs = gdfs + (gdf,)
        self._gdfs = gdfs
        if self._gdfs:
            self._gdf = pd.concat(self._gdfs, ignore_index=True)
        else:
            self._gdf = GeoDataFrame({"geometry": [], self._column: []})

        self._explore(**kwargs)

    def _load_rasters_as_images(self):
        self.raster_data = []
        n_added_images = 0
        self._show_rasters = True
        for name, value in self.rasters.items():
            data, n_added_images = self._image_collection_to_background_map(
                value,
                self.mask,
                name,
                max_images=self.max_images,
                n_added_images=n_added_images,
            )
            self.raster_data += data

    def _rasters_to_background_maps(self):
        for raster_data_dict in self.raster_data:
            try:
                arr = raster_data_dict["arr"]
            except KeyError:
                continue
            if (arr.shape) == 1:
                continue
            if hasattr(arr, "mask"):
                arr = arr.data
            if "bool" in str(arr.dtype):
                arr = np.where(arr, 1, 0)
            try:
                arr = (arr - np.min(arr)) / (np.max(arr) - np.min(arr))
            except Exception:
                pass

            label = raster_data_dict["label"]
            bounds = raster_data_dict["bounds"]
            if raster_data_dict["cmap"] is not None:
                kwargs = {"colormap": raster_data_dict["cmap"]}
            else:
                kwargs = {}
            minx, miny, maxx, maxy = bounds
            image_overlay = folium.raster_layers.ImageOverlay(
                arr,
                bounds=[[miny, minx], [maxy, maxx]],
                show=self._show_rasters,
                **kwargs,
            )
            image_overlay.layer_name = Path(label).stem
            image_overlay.add_to(self.map)

    def save(self, path: str) -> None:
        """Save the map to local disk as an html document."""
        with open(path, "w") as f:
            f.write(self.map._repr_html_())

    def _explore(self, **kwargs) -> None:
        self.kwargs = self.kwargs | kwargs

        if self._show_was_none and len([gdf for gdf in self._gdfs if len(gdf)]) > 6:
            self.show = [False] * len(self._gdfs)

        if self._is_categorical:
            self._create_categorical_map()
        else:
            self._create_continous_map()

        if self.out_path:
            with open(
                os.getcwd() + "/" + self.out_path.strip(".html") + ".html", "w"
            ) as f:
                f.write(self.map._repr_html_())
        elif self.browser:
            run_html_server(self.map._repr_html_())
        else:
            display(self.map)

    def _split_categories(self) -> None:
        new_gdfs, new_labels, new_shows = [], [], []
        for cat in self._unique_values:
            gdf = self.gdf.loc[self.gdf[self.column] == cat]
            new_gdfs.append(gdf)
            new_labels.append(cat)
            new_shows.append(self.show[0])
        self._gdfs = new_gdfs
        self._gdf = pd.concat(new_gdfs, ignore_index=True)
        self.labels = new_labels
        self.show = new_shows

    def _to_single_geom_type(self, gdf: GeoDataFrame) -> GeoDataFrame:
        gdf = clean_geoms(gdf)

        if get_geom_type(gdf) != "mixed":
            return gdf

        geom_types = gdf.geom_type.str.lower()
        mess = "Leaflet cannot render mixed geometry types well. "

        if geom_types.str.contains("collection").any():
            mess += "Exploding geometry collections. "
            gdf = make_all_singlepart(gdf)
            geom_types = gdf.geom_type.str.lower()

        if geom_types.str.contains("polygon").any():
            mess += "Keeping only polygons."
            gdf = to_single_geom_type(gdf, geom_type="polygon")

        elif geom_types.str.contains("lin").any():
            mess += "Keeping only lines."
            gdf = to_single_geom_type(gdf, geom_type="line")

        assert get_geom_type(gdf) != "mixed", gdf.geom_type.value_counts()

        warnings.warn(mess, stacklevel=1)

        return gdf

    def _update_column(self) -> None:
        if not self._gdfs:
            return
        self._is_categorical = self._check_if_categorical()
        self._fillna_if_col_is_missing()
        self._gdf = pd.concat(self._gdfs, ignore_index=True)

    def _get_bounds(
        self, gdf: GeoDataFrame
    ) -> tuple[float, float, float, float] | None:
        if not len(gdf) or all(x is None for x in gdf.total_bounds):
            try:
                return get_total_bounds([x["bounds"] for x in self.raster_data])
            except Exception:
                return None

        return gdf.total_bounds

    def _create_categorical_map(self) -> None:
        self._make_categories_colors_dict()
        if self._gdf is not None and len(self._gdf):
            self._fix_nans()
            gdf = self._prepare_gdf_for_map(self._gdf)
        else:
            gdf = GeoDataFrame({"geometry": [], self._column: []})

        self._load_rasters_as_images()

        bounds = self._get_bounds(gdf)

        if bounds is None:
            self.map = None
            return
        self.map = self._make_folium_map(
            bounds=bounds,
            max_zoom=self.max_zoom,
            popup=self.popup,
            prefer_canvas=self.prefer_canvas,
            **self.kwargs,
        )

        for gdf, label, show in zip(self._gdfs, self.labels, self.show, strict=True):
            if not len(gdf):
                continue

            gdf = self._to_single_geom_type(make_all_singlepart(gdf))

            if not len(gdf):
                continue

            gdf.geometry = shapely.force_2d(gdf.geometry.values)
            gdf = self._prepare_gdf_for_map(gdf)

            gjs = self._make_geojson(
                gdf,
                show=show,
                color=gdf["color"],
                tooltip=self._tooltip_cols(gdf),
                popup=self.popup,
                **{
                    key: value
                    for key, value in self.kwargs.items()
                    if key not in ["title", "tiles"]
                },
            )
            gjs.layer_name = label

            gjs.add_to(self.map)

        self._rasters_to_background_maps()

        _categorical_legend(
            self.map,
            self._column,
            self._categories_colors_dict.keys(),
            self._categories_colors_dict.values(),
        )

        self.map.add_child(folium.LayerControl())

    def _add_tiles(
        self, mapobj: folium.Map, tiles: list[str, xyzservices.TileProvider]
    ) -> None:
        for tile in tiles:
            to_tile(tile, max_zoom=self.max_zoom).add_to(mapobj)

    def _create_continous_map(self):
        self._prepare_continous_map()
        if self.scheme:
            classified = self._classify_from_bins(self._gdf, bins=self.bins)
            classified_sequential = self._push_classification(classified)
            n_colors = len(np.unique(classified_sequential)) - any(self._nan_idx)
            unique_colors = self._get_continous_colors(n=n_colors)

        self._load_rasters_as_images()

        gdf = self._prepare_gdf_for_map(self._gdf)
        bounds = self._get_bounds(gdf)
        if bounds is None:
            self.map = None
            return
        self.map = self._make_folium_map(
            bounds=bounds,
            max_zoom=self.max_zoom,
            popup=self.popup,
            prefer_canvas=self.prefer_canvas,
            **self.kwargs,
        )

        colorbar = bc.colormap.StepColormap(
            unique_colors,
            vmin=self._gdf[self._column].min(),
            vmax=self._gdf[self._column].max(),
            caption=self._column,
            index=self.bins,
        )

        for gdf, label, show in zip(self._gdfs, self.labels, self.show, strict=True):
            if not len(gdf):
                continue

            gdf = self._to_single_geom_type(make_all_singlepart(gdf))

            if not len(gdf):
                continue

            gdf.geometry = shapely.force_2d(gdf.geometry.values)
            gdf = self._prepare_gdf_for_map(gdf)

            classified = self._classify_from_bins(gdf, bins=self.bins)
            try:
                colorarray = unique_colors[classified]
            except IndexError:
                classified[classified > 0] = classified[classified > 0] - 1
                colorarray = unique_colors[classified]

            gjs = self._make_geojson(
                gdf,
                show=show,
                color=colorarray,
                tooltip=self._tooltip_cols(gdf),
                popup=self.popup,
                prefer_canvas=self.prefer_canvas,
                **{
                    key: value
                    for key, value in self.kwargs.items()
                    if key not in ["title", "tiles"]
                },
            )

            gjs.layer_name = label

            gjs.add_to(self.map)

        self._rasters_to_background_maps()

        self.map.add_child(colorbar)
        self.map.add_child(folium.LayerControl())

    def _tooltip_cols(self, gdf: GeoDataFrame) -> list:
        if "tooltip" in self.kwargs:
            tooltip = self.kwargs.pop("tooltip")
            return tooltip
        return [col for col in gdf.columns if col not in COLS_TO_DROP]

    @staticmethod
    def _prepare_gdf_for_map(gdf: GeoDataFrame) -> GeoDataFrame:
        if isinstance(gdf, GeoSeries):
            gdf = gdf.to_frame("geometry")

        # convert LinearRing to LineString
        rings_mask = gdf.geom_type == "LinearRing"
        if rings_mask.any():
            gdf.geometry[rings_mask] = gdf.geometry[rings_mask].apply(
                lambda g: LineString(g)
            )

        if gdf.crs is not None and not gdf.crs.equals(4326):
            gdf = gdf.to_crs(4326)

        return gdf

    def _make_folium_map(
        self,
        bounds: tuple[float, float, float, float],
        attr: Any = None,
        tiles: Any = None,
        width: str = "100%",
        height: str = "100%",
        control_scale: bool = True,
        map_kwds: dict | None = None,
        **kwargs,
    ):
        if not map_kwds:
            map_kwds = {}

        if tiles is None:
            tiles = self.tiles

        # create folium.Map object
        # Get bounds to specify location and map extent
        location = kwargs.pop("location", None)
        if location is None:
            x = mean([bounds[0], bounds[2]])
            y = mean([bounds[1], bounds[3]])
            location = (y, x)
            if "zoom_start" in kwargs.keys():
                fit = False
            else:
                fit = True
        else:
            fit = False

        # get a subset of kwargs to be passed to folium.Map
        for i in _MAP_KWARGS:
            if i in map_kwds:
                raise ValueError(
                    f"'{i}' cannot be specified in 'map_kwds'. "
                    f"Use the '{i}={map_kwds[i]}' argument instead."
                )
        map_kwds = {
            **map_kwds,
            **{i: kwargs[i] for i in kwargs.keys() if i in _MAP_KWARGS},
        }
        map_kwds["min_zoom"] = 0
        map_kwds["max_zoom"] = kwargs.get("max_zoom", self.max_zoom)

        if isinstance(tiles, (list, tuple)):
            default_tile, *more_tiles = tiles
        else:
            default_tile, more_tiles = tiles, []

        default_tile = to_tile(default_tile, max_zoom=self.max_zoom)

        if isinstance(default_tile, xyzservices.TileProvider):
            attr = attr if attr else default_tile.html_attribution
            default_tile = default_tile.build_url(scale_factor="{r}")

        m = folium.Map(
            location=location,
            control_scale=control_scale,
            tiles=default_tile,
            attr=attr,
            width=width,
            height=height,
            **map_kwds,
        )

        self._add_tiles(m, more_tiles)

        if self.measure_control:
            MeasureControlFix(
                primary_length_unit="m",
                secondary_length_unit="kilometers",
                primary_area_unit="sqmeters",
                secondary_area_unit="sqkilometers",
                position="bottomleft",
                capture_z_index=False,
                thousands_sep="",
                units={
                    "m": {"factor": 1, "display": "m", "decimals": self.decimals},
                },
            ).add_to(m)

        plugins.Fullscreen(
            position="topleft",
            title="Expand me",
            title_cancel="Exit me",
            force_separate_button=True,
        ).add_to(m)

        plugins.MousePosition(
            position="bottomleft",
            separator=", ",
            empty_string="NaN",
            lng_first=True,
            num_digits=8,
        ).add_to(m)

        if self.geocoder:
            plugins.Geocoder(position="topright").add_to(m)

        # fit bounds to get a proper zoom level
        if fit and "zoom_start" not in kwargs:
            m.fit_bounds([[bounds[1], bounds[0]], [bounds[3], bounds[2]]])

        if self.text:
            style = bc.element.MacroElement()
            style._template = bc.element.Template(get_textbox(self.text))
            m.get_root().add_child(style)
            # folium.LayerControl(collapsed=False).add_to(m)

        return m

    def _make_geojson(
        self,
        df: GeoDataFrame,
        show: bool,
        color: str | None = None,
        tooltip: bool = True,
        popup: bool = False,
        highlight: bool = True,
        marker_type: str | None = None,
        marker_kwds: dict | None = None,
        style_kwds: dict | None = None,
        highlight_kwds: dict | None = None,
        tooltip_kwds: dict | None = None,
        popup_kwds: dict | None = None,
        map_kwds: dict | None = None,
        **kwargs,
    ) -> folium.GeoJson:

        marker_kwds = marker_kwds or {}
        style_kwds = style_kwds or {}
        highlight_kwds = highlight_kwds or {}
        tooltip_kwds = tooltip_kwds or {}
        popup_kwds = popup_kwds or {}
        map_kwds = map_kwds or {}

        gdf = df.copy()

        # convert LinearRing to LineString
        rings_mask = gdf.geom_type == "LinearRing"
        if rings_mask.any():
            gdf.geometry[rings_mask] = gdf.geometry[rings_mask].apply(
                lambda g: LineString(g)
            )

        if gdf.crs is None:
            pass
            # kwargs["crs"] = "Simple"
        elif not gdf.crs.equals(4326):
            gdf = gdf.to_crs(4326)

        # get a subset of kwargs to be passed to folium.Map
        for i in _MAP_KWARGS:
            if i in map_kwds:
                raise ValueError(
                    f"'{i}' cannot be specified in 'map_kwds'. "
                    f"Use the '{i}={map_kwds[i]}' argument instead."
                )
        map_kwds = {
            **map_kwds,
            **{i: kwargs[i] for i in kwargs.keys() if i in _MAP_KWARGS},
        }

        for map_kwd in _MAP_KWARGS:
            kwargs.pop(map_kwd, None)

        # set default style
        if "fillOpacity" not in style_kwds:
            style_kwds["fillOpacity"] = 0.5
        if "weight" not in style_kwds:
            style_kwds["weight"] = 2
        if "style_function" in style_kwds:
            style_kwds_function = style_kwds["style_function"]
            if not callable(style_kwds_function):
                raise ValueError("'style_function' has to be a callable")
            style_kwds.pop("style_function")
        else:

            def _no_style(x):
                return {}

            style_kwds_function = _no_style

        gdf["__folium_color"] = color

        stroke_color = style_kwds.pop("color", None)
        if not stroke_color:

            def _style_column(x):
                base_style = {
                    "fillColor": x["properties"]["__folium_color"],
                    "color": x["properties"]["__folium_color"],
                    **style_kwds,
                }
                return {
                    **base_style,
                    **style_kwds_function(x),
                }

            style_function = _style_column

        if highlight:
            if "fillOpacity" not in highlight_kwds:
                highlight_kwds["fillOpacity"] = 0.75

            def _style_highlight(x):
                return {**highlight_kwds}

            highlight_function = _style_highlight
        else:
            highlight_function = None

        # define default for points
        if marker_type is None:
            marker_type = "circle_marker"

        marker = marker_type
        if isinstance(marker_type, str):
            if marker_type == "marker":
                marker = folium.Marker(**marker_kwds)
            elif marker_type == "circle":
                marker = folium.Circle(**marker_kwds)
            elif marker_type == "circle_marker":
                marker_kwds["radius"] = marker_kwds.get("radius", 2)
                marker_kwds["fill"] = marker_kwds.get("fill", True)
                marker = folium.CircleMarker(**marker_kwds)
            else:
                raise ValueError(
                    "Only 'marker', 'circle', and 'circle_marker' are "
                    "supported as marker values"
                )

        # prepare tooltip and popup
        if isinstance(gdf, GeoDataFrame):
            # add named index to the tooltip
            if gdf.index.name is not None:
                gdf = gdf.reset_index()
            # specify fields to show in the tooltip
            tooltip = _tooltip_popup("tooltip", tooltip, gdf, **tooltip_kwds)
            popup = _tooltip_popup("popup", popup, gdf, **popup_kwds)
        else:
            tooltip = None
            popup = None

        gdf_as_json = gdf.__geo_interface__

        return folium.GeoJson(
            gdf_as_json,
            tooltip=tooltip,
            popup=popup,
            marker=marker,
            style_function=style_function,
            highlight_function=highlight_function,
            smooth_factor=self.smooth_factor,
            show=show,
            **kwargs,
        )

    def _image_collection_to_background_map(
        self,
        image_collection: ImageCollection | Image | Band,
        mask: Any | None,
        name: str,
        max_images: int,
        n_added_images: int,
        rbg_bands: list[str] = (["B02", "B03", "B04"], ["B2", "B3", "B4"]),
    ) -> tuple[list[dict], int]:
        out = []

        if all(isinstance(x, str) for x in rbg_bands):
            rbg_bands = (rbg_bands,)

        if isinstance(image_collection, ImageCollection):
            images = image_collection.images
            name = None
        elif isinstance(image_collection, Image):
            img = image_collection
            if not _intersects_if_not_none_or_empty(
                mask, img.bounds
            ):  # is not None and not to_shapely(mask).intersects(
                #     to_shapely(img.bounds)
                # ):
                return out, n_added_images

            if len(img) == 1:
                band = next(iter(img))
                raster_data_dict = {}
                out.append(raster_data_dict)
                name = _determine_label(band, name, out, n_added_images)
                _single_band_to_arr(band, mask, name, raster_data_dict)
                n_added_images += 1
                return out, n_added_images
            elif len(img) < 3:
                raster_data_dict = {}
                out.append(raster_data_dict)
                for band in img:
                    name = _determine_label(band, None, out, n_added_images)
                    _single_band_to_arr(band, mask, name, raster_data_dict)
                    n_added_images += 1
                return out, n_added_images
            else:
                images = [image_collection]

        elif isinstance(image_collection, Band):
            band = image_collection

            if not _intersects_if_not_none_or_empty(
                mask, band.bounds
            ):  # mask is not None and not to_shapely(mask).intersects(
                #     to_shapely(band.bounds)
                # ):
                return out, n_added_images

            raster_data_dict = {}
            out.append(raster_data_dict)
            _single_band_to_arr(band, mask, name, raster_data_dict)
            return out, n_added_images

        else:
            raise TypeError(type(image_collection))

        if max(len(out), len(images)) + n_added_images > max_images:
            warnings.warn(
                f"Showing only a sample of {max_images}. Set 'max_images.", stacklevel=1
            )
            self._show_rasters = False
            random.shuffle(images)

            images = images[: (max_images - n_added_images)]
            images = (
                list(sorted([img for img in images if img.date is not None]))
                + sorted(
                    [
                        img
                        for img in images
                        if img.date is None and img.path is not None
                    ],
                    key=lambda x: x.path,
                )
                + [img for img in images if img.date is None and img.path is None]
            )

        for image in images:

            if not _intersects_if_not_none_or_empty(
                mask, image.bounds
            ):  # mask is not None and not to_shapely(mask).intersects(
                #     to_shapely(image.bounds)
                # ):
                continue

            raster_data_dict = {}
            out.append(raster_data_dict)

            if len(image) < 3:
                for band in image:
                    name = _determine_label(band, None, out, n_added_images)
                    _single_band_to_arr(band, mask, name, raster_data_dict)
                    n_added_images += 1
                continue

            for red, blue, green in rbg_bands:
                try:
                    red_band = image[red].load(indexes=1, bounds=mask)
                except KeyError:
                    continue
                try:
                    blue_band = image[blue].load(indexes=1, bounds=mask)
                except KeyError:
                    continue
                try:
                    green_band = image[green].load(indexes=1, bounds=mask)
                except KeyError:
                    continue
                break

            crs = red_band.crs
            bounds = to_bbox(to_gdf(red_band.bounds, crs).to_crs(4326))

            red_band = red_band.values
            blue_band = blue_band.values
            green_band = green_band.values

            if (
                red_band.shape[0] == 0
                or blue_band.shape[0] == 0
                or green_band.shape[0] == 0
            ):
                continue

            # to 3d array in shape (x, y, 3)
            rbg_image = np.stack([red_band, blue_band, green_band], axis=2)

            raster_data_dict["arr"] = rbg_image
            raster_data_dict["bounds"] = bounds
            raster_data_dict["cmap"] = None
            raster_data_dict["label"] = _determine_label(
                image, name, out, n_added_images
            )

            n_added_images += 1

        return out, n_added_images


def _tooltip_popup(
    type_: str, fields: Any, gdf: GeoDataFrame, **kwargs
) -> folium.GeoJsonTooltip | folium.GeoJsonPopup:
    """Get tooltip or popup."""
    # specify fields to show in the tooltip
    if fields is False or fields is None or fields == 0:
        return None
    else:
        if fields is True:
            fields = gdf.columns.drop(gdf.geometry.name).to_list()
        elif isinstance(fields, int):
            fields = gdf.columns.drop(gdf.geometry.name).to_list()[:fields]
        elif isinstance(fields, str):
            fields = [fields]

    for field in ["__plottable_column", "__folium_color"]:
        if field in fields:
            fields.remove(field)

    # Cast fields to str
    fields = list(map(str, fields))
    if type_ == "tooltip":
        return folium.GeoJsonTooltip(fields, **kwargs)
    elif type_ == "popup":
        return folium.GeoJsonPopup(fields, **kwargs)


def _intersects_if_not_none_or_empty(obj: Any, other: Any) -> bool:
    if obj is None:
        return True
    obj = to_shapely(obj)
    if obj is None or obj.is_empty:
        return True
    return obj.intersects(to_shapely(other))


def _determine_label(
    obj: Image | Band | ImageCollection, obj_name: str | None, out: list[dict], i: int
) -> str:
    # Prefer the object's name
    if obj_name:
        # Avoid the generic label e.g. Image(1)
        does_not_have_generic_name = (
            re.sub("(\d+)", "", obj_name) != f"{obj.__class__.__name__}()"
        )
        if does_not_have_generic_name:
            return obj_name
    try:
        # Images/Bands/Collections constructed from arrays have no path stems
        if obj.name:
            name = obj.name
        else:
            name = str(obj)[:23]
    except (AttributeError, ValueError):
        name = str(obj)[:23]

    if name in [x["label"] for x in out if "label" in x]:
        name += f"_{i}"

    return name


def _categorical_legend(
    m: folium.Map, title: str, categories: list[str], colors: list[str]
) -> None:
    """Add categorical legend to a map.

    The implementation is using the code originally written by Michel Metran
    (@michelmetran) and released on GitHub
    (https://github.com/michelmetran/package_folium) under MIT license.

    Copyright (c) 2020 Michel Metran

    Parameters
    ----------
    m : folium.Map
        Existing map instance on which to draw the plot
    title : str
        title of the legend (e.g. column name)
    categories : list-like
        list of categories
    colors : list-like
        list of colors (in the same order as categories)
    """
    # Header to Add
    head = """
    {% macro header(this, kwargs) %}
    <script src="https://code.jquery.com/ui/1.12.1/jquery-ui.js"></script>
    <script>$( function() {
        $( ".maplegend" ).draggable({
            start: function (event, ui) {
                $(this).css({
                    right: "auto",
                    top: "auto",
                    bottom: "auto"
                });
            }
        });
    });
    </script>
    <style type='text/css'>
      .maplegend {
        position: absolute;
        z-index:9999;
        background-color: rgba(255, 255, 255, .8);
        border-radius: 5px;
        box-shadow: 0 0 15px rgba(0,0,0,0.2);
        padding: 10px;
        font: 12px/14px Arial, Helvetica, sans-serif;
        right: 10px;
        bottom: 20px;
      }
      .maplegend .legend-title {
        text-align: left;
        margin-bottom: 5px;
        font-weight: bold;
        }
      .maplegend .legend-scale ul {
        margin: 0;
        margin-bottom: 0px;
        padding: 0;
        float: left;
        list-style: none;
        }
      .maplegend .legend-scale ul li {
        list-style: none;
        margin-left: 0;
        line-height: 16px;
        margin-bottom: 2px;
        }
      .maplegend ul.legend-labels li span {
        display: block;
        float: left;
        height: 14px;
        width: 14px;
        margin-right: 5px;
        margin-left: 0;
        border: 0px solid #ccc;
        }
      .maplegend .legend-source {
        color: #777;
        clear: both;
        }
      .maplegend a {
        color: #777;
        }
    </style>
    {% endmacro %}
    """
    import branca as bc

    # Add CSS (on Header)
    macro = bc.element.MacroElement()
    macro._template = bc.element.Template(head)
    m.get_root().add_child(macro)

    body = f"""
    <div id='maplegend {title}' class='maplegend'>
        <div class='legend-title'>{title}</div>
        <div class='legend-scale'>
            <ul class='legend-labels'>"""

    # Loop Categories
    for label, color in zip(categories, colors, strict=True):
        body += f"""
                <li><span style='background:{color}'></span>{label}</li>"""

    body += """
            </ul>
        </div>
    </div>
    """

    # Add Body
    body = bc.element.Element(body, "legend")
    m.get_root().html.add_child(body)


def get_textbox(text: str) -> str:
    return f"""
{{% macro html(this, kwargs) %}}
<!doctype html>
<html lang="en">
  <head>
    <meta charset="utf-8">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <title>Textbox Project</title>
    <link rel="stylesheet" href="//code.jquery.com/ui/1.12.1/themes/base/jquery-ui.css">
    <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.2.1/css/all.min.css" integrity="sha512-MV7K8+y+gLIBoVD59lQIYicR65iaqukzvf/nwasF0nqhPay5w/9lJmVM2hMDcnK1OnMGCdVK+iQrJ7lzPJQd1w==" crossorigin="anonymous" referrerpolicy="no-referrer"/>
    <script src="https://code.jquery.com/jquery-1.12.4.js"></script>
    <script src="https://code.jquery.com/ui/1.12.1/jquery-ui.js"></script>

    <script>
      $( function() {{
        $( "#textbox" ).draggable({{
          start: function (event, ui) {{
            $(this).css({{
              right: "auto",
              top: "auto",
              bottom: "auto"
            }});
          }}
        }});
      }});
    </script>
  </head>

  <body>
    <div id="textbox" class="textbox">
      <div class="textbox-content">
        <p>{text}</p>
      </div>
    </div>

</body>
</html>

<style type='text/css'>
  .textbox {{
    position: absolute;
    z-index:9999;
    border-radius:4px;
    background: white;
    padding: 1px;
    font-size:11px;
    left: 20px;
    top: 50%;
    color: black;
  }}
</style>
{{% endmacro %}}
"""



import numpy as np


def ndvi(red: np.ndarray, nir: np.ndarray, padding: int = 0) -> np.ndarray:
    ndvi_values = (nir - red + padding) / (nir + red + padding)
    ndvi_values[(red + nir) == 0] = 0

    return ndvi_values



import itertools
import re
from collections.abc import Sequence

import pandas as pd

from ..io._is_dapla import is_dapla

try:
    import dapla as dp
except ImportError:
    pass

try:
    from gcsfs.core import GCSFile
except ImportError:

    class GCSFile:
        """Placeholder."""


if is_dapla():

    def _open_func(*args, **kwargs) -> GCSFile:
        return dp.FileClient.get_gcs_file_system().open(*args, **kwargs)

else:
    _open_func = open


class _RegexError(ValueError):
    pass


def _any_regex_matches(xml_file: str, regexes: tuple[str]) -> bool | None:
    n_matches = 0
    for regex in regexes:
        try:
            if bool(re.search(regex, xml_file)):
                return True
            n_matches += 1
        except (TypeError, AttributeError):
            continue

    if not n_matches:
        return None
    else:
        return False


def _get_regex_match_from_xml_in_local_dir(
    paths: list[str], regexes: str | tuple[str]
) -> str | dict[str, str]:
    for i, path in enumerate(paths):
        if ".xml" not in path:
            continue
        with _open_func(path, "rb") as file:
            filebytes: bytes = file.read()
            try:
                return _extract_regex_match_from_string(
                    filebytes.decode("utf-8"), regexes
                )
            except _RegexError as e:
                if i == len(paths) - 1:
                    raise e


def _extract_regex_match_from_string(
    xml_file: str, regexes: tuple[str | re.Pattern]
) -> str | dict[str, str]:
    if all(isinstance(x, str) for x in regexes):
        for regex in regexes:
            try:
                return re.search(regex, xml_file).group(1)
            except (TypeError, AttributeError, IndexError):
                continue
        raise _RegexError()

    out = {}
    for regex in regexes:
        try:
            matches = re.search(regex, xml_file)
            out |= matches.groupdict()
        except (TypeError, AttributeError):
            continue
    if not out:
        raise _RegexError()
    return out


def _get_regexes_matches_for_df(
    df, match_col: str, patterns: Sequence[re.Pattern]
) -> pd.DataFrame:
    if not len(df):
        return df

    non_optional_groups = list(
        set(
            itertools.chain.from_iterable(
                [_get_non_optional_groups(pat) for pat in patterns]
            )
        )
    )

    if not non_optional_groups:
        return df

    assert df.index.is_unique
    keep = []
    for pat in patterns:
        for i, row in df[match_col].items():
            matches = _get_first_group_match(pat, row)
            if all(group in matches for group in non_optional_groups):
                keep.append(i)

    return df.loc[keep]


def _get_non_optional_groups(pat: re.Pattern | str) -> list[str]:
    return [
        x
        for x in [
            _extract_group_name(group)
            for group in pat.pattern.split("\n")
            if group
            and not group.replace(" ", "").startswith("#")
            and not group.replace(" ", "").split("#")[0].endswith("?")
        ]
        if x is not None
    ]


def _extract_group_name(txt: str) -> str | None:
    try:
        return re.search(r"\(\?P<(\w+)>", txt)[1]
    except TypeError:
        return None


def _get_first_group_match(pat: re.Pattern, text: str) -> dict[str, str]:
    groups = pat.groupindex.keys()
    all_matches: dict[str, str] = {}
    for x in pat.findall(text):
        if not x:
            continue
        if isinstance(x, str):
            x = [x]
        for group, value in zip(groups, x, strict=True):
            if value and group not in all_matches:
                all_matches[group] = value
    return all_matches



config = {
    "n_jobs": 1,
}


import sgis.raster.indices as indices

from .geopandas_tools.bounds import Gridlooper
from .geopandas_tools.bounds import bounds_to_points
from .geopandas_tools.bounds import bounds_to_polygon
from .geopandas_tools.bounds import get_total_bounds
from .geopandas_tools.bounds import gridloop
from .geopandas_tools.bounds import make_grid
from .geopandas_tools.bounds import make_grid_from_bbox
from .geopandas_tools.bounds import make_ssb_grid
from .geopandas_tools.buffer_dissolve_explode import buff
from .geopandas_tools.buffer_dissolve_explode import buffdiss
from .geopandas_tools.buffer_dissolve_explode import buffdissexp
from .geopandas_tools.buffer_dissolve_explode import buffdissexp_by_cluster
from .geopandas_tools.buffer_dissolve_explode import diss
from .geopandas_tools.buffer_dissolve_explode import diss_by_cluster
from .geopandas_tools.buffer_dissolve_explode import dissexp
from .geopandas_tools.buffer_dissolve_explode import dissexp_by_cluster
from .geopandas_tools.centerlines import get_rough_centerlines
from .geopandas_tools.cleaning import coverage_clean
from .geopandas_tools.cleaning import split_and_eliminate_by_longest

# from .geopandas_tools.cleaning import split_by_neighbors
from .geopandas_tools.conversion import coordinate_array
from .geopandas_tools.conversion import from_4326
from .geopandas_tools.conversion import to_4326
from .geopandas_tools.conversion import to_bbox
from .geopandas_tools.conversion import to_gdf
from .geopandas_tools.conversion import to_geoseries
from .geopandas_tools.conversion import to_shapely
from .geopandas_tools.duplicates import get_intersections  # drop_duplicate_geometries,
from .geopandas_tools.duplicates import update_geometries  # drop_duplicate_geometries,
from .geopandas_tools.general import _rename_geometry_if
from .geopandas_tools.general import clean_clip
from .geopandas_tools.general import clean_geoms
from .geopandas_tools.general import drop_inactive_geometry_columns
from .geopandas_tools.general import get_common_crs
from .geopandas_tools.general import get_grouped_centroids
from .geopandas_tools.general import get_line_segments
from .geopandas_tools.general import make_lines_between_points
from .geopandas_tools.general import points_in_bounds
from .geopandas_tools.general import random_points
from .geopandas_tools.general import random_points_in_polygons
from .geopandas_tools.general import sort_large_first
from .geopandas_tools.general import sort_long_first
from .geopandas_tools.general import sort_short_first
from .geopandas_tools.general import sort_small_first
from .geopandas_tools.general import split_out_circles
from .geopandas_tools.general import to_lines
from .geopandas_tools.geocoding import address_to_coords
from .geopandas_tools.geocoding import address_to_gdf
from .geopandas_tools.geometry_types import get_geom_type
from .geopandas_tools.geometry_types import is_single_geom_type
from .geopandas_tools.geometry_types import make_all_singlepart
from .geopandas_tools.geometry_types import to_single_geom_type
from .geopandas_tools.neighbors import get_all_distances
from .geopandas_tools.neighbors import get_k_nearest_neighbors
from .geopandas_tools.neighbors import get_neighbor_dfs
from .geopandas_tools.neighbors import get_neighbor_indices
from .geopandas_tools.neighbors import k_nearest_neighbors
from .geopandas_tools.neighbors import sjoin_within_distance
from .geopandas_tools.overlay import clean_overlay
from .geopandas_tools.point_operations import snap_all
from .geopandas_tools.point_operations import snap_within_distance
from .geopandas_tools.polygon_operations import clean_dissexp
from .geopandas_tools.polygon_operations import close_all_holes
from .geopandas_tools.polygon_operations import close_small_holes
from .geopandas_tools.polygon_operations import close_thin_holes
from .geopandas_tools.polygon_operations import eliminate_by_largest
from .geopandas_tools.polygon_operations import eliminate_by_longest
from .geopandas_tools.polygon_operations import eliminate_by_smallest
from .geopandas_tools.polygon_operations import get_cluster_mapper
from .geopandas_tools.polygon_operations import get_gaps
from .geopandas_tools.polygon_operations import get_holes
from .geopandas_tools.polygon_operations import get_polygon_clusters
from .geopandas_tools.polygon_operations import split_polygons_by_lines
from .geopandas_tools.polygons_as_rings import PolygonsAsRings
from .geopandas_tools.sfilter import sfilter
from .geopandas_tools.sfilter import sfilter_inverse
from .geopandas_tools.sfilter import sfilter_split
from .helpers import get_object_name
from .helpers import sort_nans_last
from .io.opener import opener
from .io.read_parquet import read_parquet_url
from .maps.examine import Examine
from .maps.explore import Explore
from .maps.httpserver import run_html_server
from .maps.legend import Legend
from .maps.maps import clipmap
from .maps.maps import explore
from .maps.maps import explore_locals
from .maps.maps import qtm
from .maps.maps import samplemap
from .maps.thematicmap import ThematicMap
from .maps.tilesources import kartverket as kartverket_tiles
from .maps.tilesources import xyz as xyztiles
from .networkanalysis.closing_network_holes import close_network_holes
from .networkanalysis.closing_network_holes import close_network_holes_to_deadends
from .networkanalysis.closing_network_holes import get_k_nearest_points_for_deadends
from .networkanalysis.cutting_lines import cut_lines
from .networkanalysis.cutting_lines import cut_lines_once
from .networkanalysis.cutting_lines import split_lines_by_nearest_point
from .networkanalysis.directednetwork import make_directed_network
from .networkanalysis.directednetwork import make_directed_network_norway
from .networkanalysis.finding_isolated_networks import get_component_size
from .networkanalysis.finding_isolated_networks import get_connected_components
from .networkanalysis.network import Network
from .networkanalysis.networkanalysis import NetworkAnalysis
from .networkanalysis.networkanalysisrules import NetworkAnalysisRules
from .networkanalysis.nodes import make_edge_coords_cols
from .networkanalysis.nodes import make_edge_wkt_cols
from .networkanalysis.nodes import make_node_ids
from .networkanalysis.traveling_salesman import traveling_salesman_problem
from .parallel.parallel import Parallel
from .parallel.parallel import parallel_overlay
from .parallel.parallel import parallel_overlay_rowwise
from .parallel.parallel import parallel_sjoin
from .raster.image_collection import Band
from .raster.image_collection import Image
from .raster.image_collection import ImageCollection
from .raster.image_collection import NDVIBand
from .raster.image_collection import Sentinel2Band
from .raster.image_collection import Sentinel2CloudlessBand
from .raster.image_collection import Sentinel2CloudlessCollection
from .raster.image_collection import Sentinel2CloudlessImage
from .raster.image_collection import Sentinel2Collection
from .raster.image_collection import Sentinel2Image
from .raster.image_collection import concat_image_collections

try:
    from .io.dapla_functions import check_files
    from .io.dapla_functions import get_bounds_series
    from .io.dapla_functions import read_geopandas
    from .io.dapla_functions import write_geopandas
except ImportError:
    pass



"""Check and set geometry type."""

import numpy as np
import pandas as pd
import shapely
from geopandas import GeoDataFrame
from geopandas import GeoSeries
from geopandas.array import GeometryArray
from shapely import Geometry


def make_all_singlepart(
    gdf: GeoDataFrame | GeoSeries, index_parts: bool = False, ignore_index: bool = False
) -> GeoDataFrame | GeoSeries:
    """Make all geometries single part.

    This means doing either 0, 1 or 2 calls to the explode method.

    Args:
        gdf: GeoDataFrame
        index_parts: Defaults to False.
        ignore_index: Defaults to False.

    Returns:
        A GeoDataFrame of singlepart geometries.
    """
    # only explode if nessecary
    if index_parts or ignore_index:
        gdf = gdf.explode(index_parts=index_parts, ignore_index=ignore_index)

    while not np.all(
        np.isin(_get_geom_types(gdf), ["Polygon", "Point", "LineString", "LinearRing"])
    ):
        gdf = gdf.explode(index_parts=index_parts, ignore_index=ignore_index)

    return gdf


def to_single_geom_type(
    gdf: GeoDataFrame | GeoSeries | Geometry | GeometryArray | np.ndarray,
    geom_type: str,
    ignore_index: bool = False,
) -> GeoDataFrame | GeoSeries:
    """Returns only the specified geometry type in a GeoDataFrame or GeoSeries.

    GeometryCollections are first exploded, then only the rows with the given
    geometry_type is kept. Both multipart and singlepart geometries are kept.
    LinearRings are considered lines.

    Args:
        gdf: GeoDataFrame or GeoSeries
        geom_type: the geometry type to keep, either 'point', 'line' or 'polygon'. Both
            multi- and singlepart geometries are included.
        ignore_index: If True, the resulting axis will be labeled 0, 1, …, n - 1.
            Defaults to False.

    Returns:
        A GeoDataFrame with a single geometry type.

    Raises:
        TypeError: If incorrect gdf type.
        ValueError: If 'geom_type' is neither 'polygon', 'line' or 'point'.

    Examples:
    ---------
    First create a GeoDataFrame of mixed geometries.

    >>> from sgis import to_gdf, to_single_geom_type
    >>> from shapely.geometry import LineString, Polygon
    >>> gdf = to_gdf([
    ...     (0, 0),
    ...     LineString([(1, 1), (2, 2)]),
    ...     Polygon([(3, 3), (4, 4), (3, 4), (3, 3)])
    ...     ])
    >>> gdf
                                                geometry
    0                            POINT (0.00000 0.00000)
    1      LINESTRING (1.00000 1.00000, 2.00000 2.00000)
    2  POLYGON ((3.00000 3.00000, 4.00000 4.00000, 3....

    >>> to_single_geom_type(gdf, "line")
                                            geometry
    1  LINESTRING (1.00000 1.00000, 2.00000 2.00000)
    >>> to_single_geom_type(gdf, "polygon")
                                                geometry
    2  POLYGON ((3.00000 3.00000, 4.00000 4.00000, 3....

    Also keeps multigeometries and geometries within GeometryCollections.

    >>> gdf = gdf.dissolve()
    >>> gdf
                                                geometry
    0  GEOMETRYCOLLECTION (POINT (0.00000 0.00000), L...

    >>> to_single_geom_type(gdf, "line")
                                            geometry
    2  LINESTRING (1.00000 1.00000, 2.00000 2.00000)
    """
    if all(g not in geom_type for g in ["polygon", "line", "point"]):
        raise ValueError(
            f"Invalid geom_type {geom_type!r}. Should be 'polygon', 'line' or 'point'"
        )

    if isinstance(gdf, Geometry):
        return _shapely_to_single_geom_type(gdf, geom_type)

    if isinstance(gdf, (np.ndarray, GeometryArray)):
        arr = np.vectorize(_shapely_to_single_geom_type)(gdf, geom_type)
        return arr[~shapely.is_empty(arr)]

    if not isinstance(gdf, (GeoDataFrame, GeoSeries)):  # type: ignore [unreachable]
        raise TypeError(f"'gdf' should be GeoDataFrame or GeoSeries, got {type(gdf)}")

    # explode collections to single-typed geometries
    collections = _get_geom_types(gdf) == "GeometryCollection"
    if collections.any():
        collections = make_all_singlepart(gdf[collections], ignore_index=ignore_index)

        gdf = pd.concat([gdf, collections], ignore_index=ignore_index)

    if "poly" in geom_type:
        is_polygon = np.isin(_get_geom_types(gdf), ["Polygon", "MultiPolygon"])
        if not is_polygon.all():
            gdf = gdf.loc[is_polygon]
    elif "line" in geom_type:
        is_line = np.isin(
            _get_geom_types(gdf), ["LineString", "MultiLineString", "LinearRing"]
        )
        if not is_line.all():
            gdf = gdf.loc[is_line]
    else:
        is_point = np.isin(_get_geom_types(gdf), ["Point", "MultiPoint"])
        if not is_point.all():
            gdf = gdf.loc[is_point]

    return gdf.reset_index(drop=True) if ignore_index else gdf


def _shapely_to_single_geom_type(geom: Geometry, geom_type: str) -> Geometry:
    parts = shapely.get_parts(geom)
    return shapely.union_all(
        [part for part in parts if geom_type.lower() in part.geom_type.lower()]
    )


def get_geom_type(gdf: GeoDataFrame | GeoSeries) -> str:
    """Returns a string of the geometry type in a GeoDataFrame or GeoSeries.

    Args:
        gdf: GeoDataFrame or GeoSeries

    Returns:
        A string that is either "polygon", "line", "point", or "mixed".

    Raises:
        TypeError: If 'gdf' is not of type GeoDataFrame or GeoSeries.

    Examples:
    ---------
    >>> from sgis import to_gdf, get_geom_type
    >>> gdf = to_gdf([0, 0])
    >>> gdf
                      geometry
    0  POINT (0.00000 0.00000)
    >>> get_geom_type(gdf)
    'point'
    """
    polys = ["Polygon", "MultiPolygon", None]
    lines = ["LineString", "MultiLineString", "LinearRing", None]
    points = ["Point", "MultiPoint", None]
    if isinstance(gdf, Geometry):
        if gdf.geom_type in polys:
            return "polygon"
        if gdf.geom_type in lines:
            return "line"
        if gdf.geom_type in points:
            return "point"
        return "mixed"

    if np.all(np.isin(_get_geom_types(gdf), polys)):
        return "polygon"
    if np.all(np.isin(_get_geom_types(gdf), lines)):
        return "line"
    if np.all(np.isin(_get_geom_types(gdf), points)):
        return "point"
    return "mixed"


def is_single_geom_type(gdf: GeoDataFrame | GeoSeries) -> bool:
    """Returns True if all geometries in a GeoDataFrame are of the same type.

    The types are either polygon, line or point. Multipart and singlepart are
    considered the same type.

    Args:
        gdf: GeoDataFrame or GeoSeries

    Returns:
        True if all geometries are the same type, False if not.

    Raises:
        TypeError: If 'gdf' is not of type GeoDataFrame or GeoSeries.

    Examples:
    ---------
    >>> from sgis import to_gdf, get_geom_type
    >>> gdf = to_gdf([0, 0])
    >>> gdf
                                                geometry
    0                            POINT (0.00000 0.00000)
    >>> is_single_geom_type(gdf)
    True
    """
    return (
        np.all(np.isin(_get_geom_types(gdf), ["Polygon", "MultiPolygon"]))
        or np.all(
            np.isin(
                _get_geom_types(gdf), ["LineString", "MultiLineString", "LinearRing"]
            )
        )
        or np.all(np.isin(_get_geom_types(gdf), ["Point", "MultiPoint"]))
    )


def _get_geom_types(
    gdf: GeoDataFrame | GeoSeries | GeometryArray | np.ndarray,
) -> np.ndarray:
    try:
        return GeometryArray(gdf.geometry.values).geom_type
    except AttributeError:
        return getattr(GeometryArray(gdf, "geom_type"))



from collections.abc import Callable
from collections.abc import Sequence

import geopandas as gpd
import numpy as np
import pandas as pd
from shapely import Geometry
from shapely.geometry import Polygon

from ..helpers import get_non_numpy_func_name
from ..helpers import get_numpy_func


def _prepare_zonal(
    polygons: gpd.GeoDataFrame,
    aggfunc: str | Callable | Sequence[Callable | str],
) -> tuple[gpd.GeoDataFrame, list[Callable], list[str]]:
    polygons = polygons.reset_index(drop=True)[["geometry"]]

    if isinstance(aggfunc, str) or callable(aggfunc):
        aggfunc = [aggfunc]

    aggfunc = [f if callable(f) else get_numpy_func(f) for f in aggfunc]

    func_names = [get_non_numpy_func_name(f) for f in aggfunc]

    return polygons, aggfunc, func_names


def _make_geometry_iterrows(gdf: gpd.GeoDataFrame) -> list[tuple[int, Geometry]]:
    """Because pandas iterrows returns non-geo Series."""
    return list(gdf.geometry.items())


def _zonal_func(
    poly_iter: tuple[int, Polygon],
    cube,
    array_func: Callable,
    aggfunc: str | Callable | Sequence[Callable | str],
    func_names: list[str],
    by_date: bool,
) -> pd.DataFrame:
    cube = cube.copy()
    i, polygon = poly_iter
    if not by_date or cube.date.isna().all():
        df = _clip_and_aggregate(
            cube, polygon, array_func, aggfunc, func_names, date=None, i=i
        )
        return df if by_date else df.drop(columns="date")

    out = []

    na_date = cube[cube.date.isna()]
    df = _clip_and_aggregate(
        na_date, polygon, array_func, aggfunc, func_names, date=pd.NA, i=i
    )
    out.append(df)

    cube = cube[lambda x: x["date"].notna()]

    for dt in cube.date.unique():
        cube_date = cube[cube.date == dt]
        df = _clip_and_aggregate(
            cube_date, polygon, array_func, aggfunc, func_names, dt, i
        )
        out.append(df)

    return pd.concat(out)


def _no_overlap_df(func_names: list[str], i: int, date: str) -> pd.DataFrame:
    df = pd.DataFrame(columns=func_names, index=[i])
    df["date"] = date
    df["_no_overlap"] = 1
    return df


def _clip_and_aggregate(
    cube,
    polygon: Geometry,
    array_func: Callable,
    aggfunc: Callable,
    func_names: list[str],
    date: str,
    i: int,
) -> pd.DataFrame:
    if not len(cube):
        return _no_overlap_df(func_names, i, date)
    clipped = cube.clipmerge(polygon)
    if not len(clipped) or [arr is None for arr in clipped.arrays]:
        return _no_overlap_df(func_names, i, date)
    assert len(clipped) == 1
    array = clipped[0].array
    df = _aggregate(array, array_func, aggfunc, func_names, date, i)
    return df


def _aggregate(
    array: np.ndarray,
    array_func: Callable,
    aggfunc: Callable,
    func_names: list[str],
    date: str,
    i: int,
) -> pd.DataFrame:
    if array_func:
        array = array_func(array)
    # flat_array = array.astype(np.float64).flatten()
    flat_array = array.flatten()
    no_nans = flat_array[~np.isnan(flat_array)]
    data = {}
    for f, name in zip(aggfunc, func_names, strict=True):
        num = f(no_nans)
        data[name] = num
    df = pd.DataFrame(data, index=[i])
    df["date"] = date
    df["_no_overlap"] = 0
    return df


def _zonal_post(
    aggregated: list[pd.DataFrame],
    polygons: gpd.GeoDataFrame,
    idx_mapper: dict | pd.Series,
    idx_name: str,
    dropna: bool,
) -> pd.DataFrame:
    out = gpd.GeoDataFrame(
        pd.concat(aggregated), geometry=polygons.geometry.values, crs=polygons.crs
    ).sort_index()

    out.index = out.index.map(idx_mapper)
    out.index.name = idx_name

    if dropna:
        out = out.loc[~out.drop(columns="geometry").isna().all(axis=1)]
        return out.loc[out["_no_overlap"] != 1].drop(columns="_no_overlap")
    else:
        return out.drop(columns="_no_overlap")



try:
    import geocoder
except ImportError:
    pass
from geopandas import GeoDataFrame

from .conversion import to_gdf


def address_to_gdf(address: str, crs=4326) -> GeoDataFrame:
    """Takes an address and returns a point GeoDataFrame."""
    g = geocoder.osm(address).json
    coords = g["lng"], g["lat"]
    return to_gdf(coords, crs=4326).to_crs(crs)


def address_to_coords(address: str, crs=4326) -> tuple[float, float]:
    """Takes an address and returns a tuple of xy coordinates."""
    g = geocoder.osm(address).json
    coords = g["lng"], g["lat"]
    point = to_gdf(coords, crs=4326).to_crs(crs)
    x, y = point.geometry.iloc[0].x, point.geometry.iloc[0].y
    return x, y
